from io import BytesIO
#from cStringIO import StringIO as BytesIO

from numpy import ndarray, array, fromstring, float32, ones, empty, newaxis, savetxt, sqrt, mod, isnan, ma, hstack, log10
from reductus.dataflow.lib.exporters import exports_HDF5, exports_text

from .MetaArray import MetaArray

class FilterableMetaArray(MetaArray):
    def __new__(cls, *args, **kw):
        obj = MetaArray.__new__(cls, *args, **kw)
        #print("fma extra")
        obj.extrainfo = obj._info[-1]
        return obj

    def todict(self):
        output = { 'shape': self.shape, 'type': str(self.dtype), 'info': self.infoCopy()}
        output['data'] = self.tolist()
        return _toDictItem(output)

    def dumps(self):
        meta = { 'shape': self.shape, 'type': str(self.dtype), 'info': self.infoCopy()}
        assert isinstance(meta['info'], list)
        axstrs = []
        for ax in meta['info']:
            if 'values' in ax:
                axstrs.append(ax['values'].tostring())
                ax['values_len'] = len(axstrs[-1])
                ax['values_type'] = str(ax['values'].dtype)
                del ax['values']
        fd = BytesIO()
        fd.write(str(meta) + '\n\n')
        for ax in axstrs:
            fd.write(ax)
        fd.write(self.tostring())
        ans = fd.getvalue()
        fd.close()
        return ans

    def get_extrema(self):
        extrema = {}
        for ax in self._info:
            if 'values' in ax:
                values = array(ax['values'])
                extrema[ax['name']] = [values.min(), values.max()]
        return extrema

    @classmethod
    def loads(cls, str):
        fd = BytesIO(str)
        meta = ''
        while True:
            line = fd.readline().strip()
            if line == '':
                break
            meta += line
        meta = eval(meta)

        ## read in axis values
        for ax in meta['info']:
            if 'values_len' in ax:
                ax['values'] = fromstring(fd.read(ax['values_len']), dtype=ax['values_type'])
                del ax['values_len']
                del ax['values_type']

        subarr = fromstring(fd.read(), dtype=meta['type'])
        subarr = subarr.view(FilterableMetaArray)
        subarr.shape = meta['shape']
        subarr._info = meta['info']
        return subarr

    def use_binary(self):
        if len(self.shape) == 3:
            return True
        elif len(self.shape) == 2:
            return False
        else:
            return False

    def get_metadata(self):
        metadata = {}
        metadata.update(self.extrainfo)
        return metadata

    @exports_text(name="column")
    def to_column_text(self):
        name = self.extrainfo["friendly_name"]
        entry = self.extrainfo["entry"]
        suffix = ".ospec.dat"
        filename = "%s_%s%s" % (name, entry, suffix)
        return_value = {"name": name, "entry": entry, "file_suffix": suffix}
        if len(self.shape) == 3:
            # return gnuplottable format:
            """ export 2d data to gnuplot format """
            # grab the first counts col:
            data = self
            cols = data._info[2]['cols']
            data_cols = [col['name'] for col in cols if col['name'].startswith('counts')]

            result = []
            for colnum, col in enumerate(data_cols):
                array_out = data['Measurements':col].view(ndarray)

                dump = ""
                xlist = data._info[0]['values'].tolist()
                ylist = data._info[1]['values'].tolist()
                for ix, x in enumerate(xlist):
                    dump += "\n"
                    for iy, y in enumerate(ylist):
                        dump += "%g\t%g\t%g\n" % (x, y, array_out[ix, iy])
                result.append(dump)

            return_value["value"] =  result[0]

        elif len(self.shape) == 2:
            fid = BytesIO()
            cols = self._info[1]['cols']
            # put x axis in first:
            data_cols = [self._info[0]['name']]
            data_cols.extend([col['name'] for col in cols if not col['name'].startswith('error')])
            output_data = hstack((self._info[0]['values'][:,None], self))
            savetxt(fid, output_data, header="\t".join(data_cols))
            fid.seek(0)
            return_value["value"] = fid.read().decode()

        else:
            print("can only handle 1d or 2d data")
            return_value["value"] = ""

        return return_value

    def get_plottable(self, binary_fp=None):
        if len(self.shape) == 3:
            return self.get_plottable_2d(binary_fp)
        elif len(self.shape) == 2:
            return self.get_plottable_nd()
        else:
            print("can only handle 1d or 2d data")
            return

    def get_plottable_1d(self, binary_fp=None):
        colors = ['Blue', 'Red', 'Green', 'Yellow']
        cols = self._info[1]['cols']
        data_cols = [col['name'] for col in cols if not col['name'].startswith('error')]
        x = self._info[0]['values'].tolist()
        xlabel = self._info[0]['name']
        title = self._info[-1].get('friendly_name', "1d data")
        plottable_data = {
            'type': '1d',
            'entry': self.extrainfo.get('entry', 'entry'),
            'title': title,
            'options': {'series': []},
            'clear_existing': False,
            'data': []
        }

        for i, col in enumerate(data_cols):
            y = self['Measurements':col].tolist()
            #error_col = next((i for i in xrange(len(cols)) if cols[i]['name'] == ('error_'+col)), -1)
            #if error_col > 0:
            #    yerror = self['Measurements':'error_'+col].tolist()
            #else:
            #    yerror = sqrt(abs(self['Measurements':col])).tolist()
            series_data = [[xx,yy] for xx, yy in zip(x,y)]
            plottable_data['data'].append(series_data)
            plottable_data['options']['series'].append({'label': col})

        return [plottable_data]

    def get_plottable_nd(self, binary_fp=None):
        colors = ['Blue', 'Red', 'Green', 'Yellow']
        cols = self._info[1]['cols']
        data_cols = [col['name'] for col in cols if not col['name'].startswith('error')]
        column_listing = dict([(d, {"label": d}) for d in data_cols])
        x = self._info[0]['values'].tolist()
        xlabel = self._info[0]['name']
        column_listing[xlabel] = {"label": xlabel}
        plottable_data = {
            'type': 'nd',
            'title': self.extrainfo['path'],
            'name': self._info[-1].get('filename', '1d data'),
            'entry': self._info[-1].get('entry', ''),
            'columns': column_listing,
            "options": {
                "series": [ 
                    {
                        'label': self._info[-1].get('friendly_name', '1d data'),
                        'color': 'Red',
                        'style': 'line',
                    }
                ],
                'xcol': xlabel,
                'ycol': cols[0]['name'],
                "errorbar_width": 0
            },
            'datas': {xlabel: {'values': x}},
            'clear_existing': False,
        }

        for i, col in enumerate(data_cols):
            y = self['Measurements':col].tolist()
            error_col = next((i for i in range(len(cols)) if cols[i]['name'] == ('error_'+col)), -1)
            series_y = {'values': y}
            if error_col > 0:
                series_y['errorbars'] = self['Measurements':'error_'+col].tolist()
            plottable_data['datas'][col] = series_y

        return plottable_data

    def get_plottable_2d(self, binary_fp=None):
        # grab the first counts col:
        cols = self._info[2]['cols']
        counts_col = next(iter([col['name'] for col in cols if col['name'].startswith('counts')]))
        mon_col = next(iter([col['name'] for col in cols if col['name'].startswith('monitor')]))
        array_out = self['Measurements':counts_col].view(ndarray) / self['Measurements':mon_col].view(ndarray) 

        dump = {'entry': self.extrainfo.get('entry', 'entry')}
        if binary_fp is not None:
            # use lookup to get binary value
            z = [[0,0]]
            dump['binary_fp'] = binary_fp + ":" + str(colnum)
        else: # use the old way
            af = array_out.ravel('C')
            z = [ma.masked_array(af, mask=isnan(af)).tolist(fill_value=None)]

        #zbin_base64 = base64.b64encode(array_out.tostring())
        #z = [arr[:, 0].tolist() for arr in self]
        dims = {}
        # can't display zeros effectively in log... set zmin to smallest non-zero

        lowest = 1e-10
        non_zeros = array_out[array_out > lowest]
        if len(non_zeros) > 0:
            dims['zmin'] = float(non_zeros.min())
            dims['zmax'] = float(non_zeros.max())
        else:
            dims['zmin'] = float(lowest)
            dims['zmax'] = float(lowest)

        #dims['zmin'] = array_out.min()
        #dims['zmax'] = array_out.max()
        axis = ['x', 'y']
        for index, label in enumerate(axis):
            arr = self._info[index]['values']
            dims[axis[index] + 'min'] = float(arr[0])
            dims[axis[index] + 'max'] = float(arr[-1])
            dims[axis[index] + 'dim'] = len(arr)
        xlabel = self._info[0]['name']
        ylabel = self._info[1]['name']
        zlabel = counts_col
        #zlabel = self._info[2]['cols'][0]['name']
        title = 'MAGIK data' # That's creative enough, right?
        plot_type = '2d'
        transform = 'log' # this is nice by default
        dump.update( dict(type=plot_type, z=z, title=title, dims=dims,
                    xlabel=xlabel, ylabel=ylabel, zlabel=zlabel,
                    transform=transform) )
        return dump

    def get_plottable_binary(self):
        cols = self._info[2]['cols']
        data_cols = [col['name'] for col in cols if col['name'].startswith('counts')]

        result = []
        for col in data_cols:
            # output in column-major order, since first index is "x",
            # and we want to traverse that axis first.
            array_out = self['Measurements':col].view(ndarray)
            array_out = array_out.ravel('F')
            result.append(array_out.astype(float32).tostring())

        return result

    def get_csv(self):
        if len(self.shape) == 3:
            num_cols = self.shape[2]
            new_array = empty((self.shape[0] * self.shape[1], num_cols + 2))
            new_array[:,0] = (self._info[0]['values'][:,newaxis] * ones((self.shape[0], self.shape[1]))).ravel()
            new_array[:,1] = (self._info[1]['values'][newaxis,:] * ones((self.shape[0], self.shape[1]))).ravel()
            data_names = []
            data_names.append(self._info[0]['name']) # xlabel
            data_names.append(self._info[1]['name']) # ylabel

            for i in range(num_cols):
                new_array[:,i+2] = self[:,:,i].view(ndarray).ravel()
                data_names.append(self._info[2]['cols'][i]['name'])

            outstr = BytesIO()
            outstr.write('#' + '\t'.join(data_names) + '\n')
            savetxt(outstr, new_array, delimiter='\t', newline='\n')

            outstr.seek(0)
            return_val = outstr.read()
            outstr.close()

            return return_val

        elif len(self.shape) == 2:
            num_cols = self.shape[1]
            new_array = empty((self.shape[0], num_cols + 1))
            new_array[:,0] = (self._info[0]['values'])
            data_names = []
            data_names.append(self._info[0]['name']) # xlabel

            for i in range(num_cols):
                new_array[:,i+1] = self[:,i].view(ndarray)
                data_names.append(self._info[1]['cols'][i]['name'])

            outstr = BytesIO()
            outstr.write('#' + '\t'.join(data_names) + '\n')
            savetxt(outstr, new_array, delimiter='\t')

            outstr.seek(0)
            return_val = outstr.read()
            outstr.close()
            return return_val

        else:
            print("can only handle 1d or 2d data")
            return

#    def get_plottable_new(self):
#        array_out = self['Measurements':'counts']
#        z = {'png': base64.b64encode(array_to_png(array_out, colormap='jet')),
#             'data': array_out.tolist()}
#        dims = {}
#        dims['zmin'] = array_out.min()
#        dims['zmax'] = array_out.max()
#        axis = ['x', 'y']
#        for index, label in enumerate(axis):
#            arr = self._info[index]['values']
#            dims[axis[index] + 'min'] = amin(arr)
#            dims[axis[index] + 'max'] = amax(arr)
#            dims[axis[index] + 'dim'] = alen(arr)
#            dims['d' + axis[index]] = arr[1] - arr[0]
#        xlabel = self._info[0]['name']
#        ylabel = self._info[1]['name']
#        zlabel = self._info[2]['cols'][0]['name']
#        title = 'AND/R data' # That's creative enough, right?
#        type = '2d_image'
#        dump = dict(type=type, z=z, title=title, dims=dims, xlabel=xlabel, ylabel=ylabel, zlabel=zlabel)
#        res = simplejson.dumps(dump, sort_keys=True)
#        return res

import datetime
from collections import OrderedDict
import numpy as np

def _toDictItem(obj, convert_bytes=False):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, list):
        obj = [_toDictItem(a, convert_bytes=convert_bytes) for a in obj]
    elif isinstance(obj, dict):
        obj = OrderedDict([(k, _toDictItem(v, convert_bytes=convert_bytes)) for k, v in obj.items()])
    elif isinstance(obj, bytes) and convert_bytes == True:
        obj = obj.decode()
    return obj


from os.path import basename
from io import BytesIO

from reductus.dataflow.fetch import url_get


def load_from_string(filename, data, entries=None, loader=None):
    """
    Load a nexus file from a string, e.g., as returned from url.read().
    """
    with BytesIO(data) as fd:
        entries = loader(filename, fd, entries=entries)
    return entries

def url_load(fileinfo, check_timestamps=True, loader=None):
    path, entries = fileinfo['path'], fileinfo.get('entries', None)
    filename = basename(path)
    content = url_get(fileinfo, mtime_check=check_timestamps)
    if loader is not None:
        return load_from_string(filename, content, entries=entries,
                                loader=loader)
    elif filename.endswith('.raw') or filename.endswith('.ras') or filename.endswith('.xrdml'):
        from . import xrawref
        return load_from_string(filename, content, entries=entries,
                                loader=xrawref.load_entries)
    elif filename.endswith('.nxs.cdr'):
        from . import candor
        return load_from_string(filename, content, entries=entries,
                                loader=candor.load_entries)
    else:
        from . import nexusref
        return load_from_string(filename, content, entries=entries,
                                loader=nexusref.load_entries)

def url_load_list(files=None, check_timestamps=True, loader=None):
    if files is None:
        return []
    result = [
        entry
        for fileinfo in files
        for entry in url_load(
            fileinfo, check_timestamps=check_timestamps, loader=loader,
            )
        ]
    return result

def setup_fetch():
    #from web_gui import default_config
    from reductus.dataflow.cache import set_test_cache
    from reductus.dataflow import fetch

    set_test_cache()
    fetch.DATA_SOURCES = [
        {
            "name": "file",
            "url": "file:///",
            "start_path": "",
        },
        {
            "name": "https",
            "url": "https://",
            "start_path": "",
        },
        {
            "name": "http",
            "url": "http://",
            "start_path": "",
        },
        {
            "name": "nice",
            "url": "file:///",
            "start_path": "usr/local/nice/server_data/experiments",
        },
        {
            "name": "DOI",
            "DOI": "10.18434/T4201B",
        },
        {
            "name": "ncnr",
            "url": "https://ncnr.nist.gov/pub/",
            "start_path": "",
        },
        {
            "name": "charlotte",
            "url": "http://charlotte.ncnr.nist.gov/pub",
            "start_path": "",
        },
    ]

def fetch_uri(uri, loader=None):
    import os.path

    if ':' not in uri:
        # some sort of filename...
        source, path = 'file', os.path.realpath(os.path.expanduser(uri))
    else:
        source, path = uri.split('://')
    entries = url_load(
        {'source': source, 'path': path},
        check_timestamps=False,
        loader=loader,
        )
    return entries



"""
Generate module definitions from function declarations.

:func:`make_modules` defines the modules available for the instrument given
a list of actions.  The doc strings of the actions define the interface.

:func:`make_template` defines a convenient syntax for creating templates
from a script.
"""
import sys
import inspect
import re

from numpy import inf

from .anno_exc import annotate_exception
from .core import Module, Template
from .rst2html import rst2html

IS_PY3 = sys.version_info[0] >= 3

# Decorators to tag action methods for an instrument

def cache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = True
    return action

def nocache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = False
    return action

def module(tag=""):
    """
    Decorator adds *group=tag* as an attribute to the function.

    This marks a method as a reduction step to be included in the list
    of reduction steps. If *tag* is set then the action will be placed in a
    submenu with that label.  If used as a bare function then defaults
    to a tag of "" for the top-level menu.

    For example, to register *action*::

        @module
        def action(data, par='default'):
            ...

    To register action in the *viz* submenu use::

        @module("viz")
        def viz_action(data):
            ...

    Actions can be retrieved from a module using :func:`get_modules`.
    """
    # Called as @module
    if callable(tag):
        tag.group = ""
        return tag

    # Called as @module("tag")
    def wrapper(fn):
        fn.group = tag
        return fn
    return wrapper

# From unutbu and Glenn Maynard
# https://stackoverflow.com/questions/13503079/how-to-create-a-copy-of-a-python-function/13503277#13503277
# https://stackoverflow.com/a/6528148/190597
def copy_func(f):
    """Make a copy of a function, including its attributes"""
    import types
    import functools
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g

def copy_module(f, new_name, old_type, new_type, tag=None):
    """
    Copy a dataflow module replacing name with *new_name* and *old_type*
    with *new_type*.

    For example::

        from reductus.dataflow.automod import copy_module
        # Note: do not load symbols from steps directly into the file scope
        # or they will be defined twice as reduction modules.
        from reductus.reflred import steps

        candor_join = copy_module(
            steps.join, "candor_join", "refldata", "candordata", tag="candor")
    """
    g = copy_func(f)
    g.__name__ = new_name
    g.__doc__ = re.sub(
        r"\({old_type}([^a-zA-Z_)]*)\)".format(old_type=old_type),
        r"({new_type}\1)".format(new_type=new_type),
        f.__doc__)
    if tag is not None:
        g.group = tag
    return g

def get_modules(module, grouped=False, sorted=True):
    """
    Retrieve @module actions from a python module.

    If *grouped*, return modules grouped by action tag.

    If *sorted*, sort the actions by name attribute, otherwise they appear
    in the original order in the module.
    """
    actions = [
        fn for name in dir(module)
        for fn in [getattr(module, name)]
        if hasattr(fn, 'group')
    ]
    if sorted:
        actions.sort(key=lambda fn: fn.__name__)
    if grouped:
        from collections import defaultdict
        groups = defaultdict(list)
        for fn in actions:
            groups[fn.action].append(fn)
        return groups
    return actions

def make_template(name, description, diagram, instrument, version):
    """
    Convert a diagram into a template.

    A diagram is a list of action steps with configuration for each action
    and links between the steps.  Each step is indicated by a
    (string, dictionary) pair, with the string giving the action and
    the dictionary giving the configuration value for each field and
    input terminal in the action.  Links are given as 'source.terminal',
    where source is '-' to link to the action in the previous step.
    Source can also be the name of an action.  In cases where the action
    is duplicated (e.g., join), use "action => source" instead to give
    the source a unique name.  Use ',' to separate sources for multiple
    inputs sources for the same terminal.

    For example::

        diagram = [
            ["ncnr_load", {}],
            ["mark_intent", {"data": "-.output", "intent": "auto"}],
            ["group_by_intent => split", {"data": "-.output"}],

            ["join => spec", {"data": "split.specular"}],
            ["join => backp",  {"data": "split.backp"},],
            ["join => backm",  {"data": "split.backm"}],
            ["join => slit",  {"data": "split.slit", "tolerance": 0.0001}],

            ["subtract_background", {
                "data": "spec.output",
                "backp": "backp.output",
                "backm": "backm.output"}],
            ["divide_intensity",  {"data": "-.output", "base": "slit.output"}],
        ]

    Note that spacing is optional in action and source values.

    The diagram is a directed acyclic graph, and so forward references aren't
    required and aren't accepted.  Ambiguous references aren't identified, and
    if you refer to an action, the most recent action will be used as the
    source, not the first action.
    """
    def err(s):
        return "%s in step %d: [%r, %r]"%(s, target_index, action, config)
    def target_err(s):
        return "%s in step %d: [%r, %r]" %(
            s, target_index, action, {target_terminal:s})

    module_handles = []
    modules = []
    wires = []
    refs = {}
    for target_index, (action, config) in enumerate(diagram):

        # Split "name => ref" into "name", "ref"
        parts = [w.strip() for w in action.split('=>')]
        if len(parts) == 1:
            module_name = parts[0]
            refs[module_name] = target_index
        elif len(parts) == 2:
            module_name, ref = parts
            if ref in refs:
                raise ValueError(err("redefining source reference"))
            refs[ref] = target_index
        else:
            raise ValueError(err("too many source references"))

        try:
            module = instrument.get_module_by_id(module_name)
        except KeyError:
            raise ValueError(err("action not defined for %s"%instrument.id))
        module_handles.append(module)

        # Check that we are not setting config info for the output terminals.
        for target in module.outputs:
            target_terminal = target["id"]
            if target_terminal in config:
                raise ValueError("link should not be set on output terminal %r"
                                 %target_terminal)

        # Scan through the input terminals, and for each input terminal
        # convert the config information for that terminal into a wire.
        config = config.copy()
        for target in module.inputs:
            # retrieve the target terminal, and its configuration
            target_terminal = target["id"]
            config_value = config.pop(target_terminal, None)

            # check if it is required
            if config_value is None:
                if target["required"]:
                    raise ValueError(target_err("input terminal requires link"))
                continue

            # check for multiple links
            parts = [w.strip() for w in config_value.split(',')]
            if len(parts) > 1 and not target["multiple"]:
                raise ValueError(target_err("only one input allowed"))

            # for each given link
            for p in parts:
                # make sure it is source.terminal
                try:
                    source_id, source_terminal = p.split('.')
                except:
                    expected = {target_terminal: 'source.terminal'}
                    raise ValueError(target_err("expected %r"%(expected,)))

                # interpret source
                if source_id == '-':
                    if target_index == 0:
                        raise ValueError(target_err("no prior step"))
                    source_index = target_index-1
                elif source_id in refs:
                    source_index = refs[source_id]
                else:
                    raise ValueError(target_err("source does not exist"))

                # retrieve source terminal info
                source_module = module_handles[source_index]
                try:
                    source = source_module.get_terminal_by_id(source_terminal)
                except KeyError:
                    raise ValueError(target_err("terminal %r does not exist"
                                                % source_terminal))

                # make sure that source is an output terminal of the correct
                # data type
                if source not in source_module.outputs:
                    raise ValueError(target_err("terminal %r is not an output terminal"
                                                % source_terminal))
                if target["datatype"] != source["datatype"]:
                    raise ValueError(target_err("data types don't match"))

                # all is good; create a wire and add it to the list of wires
                wires.append({"source": [source_index, source_terminal],
                              "target": [target_index, target_terminal]})

        # TODO: validate field inputs in template, or at least check the type
        # check that the remaining config info refers to fields in the
        # module
        fields = set(p["id"] for p in module.fields)
        for k, v in config.items():
            if k not in fields:
                raise ValueError(err("unknown config field %r"%k))

        # Add the module to the list of modules in the template
        modules.append({"module": module.id,
                        "version": module.version,
                        "config": config})

    return Template(name=name, description=description, version=version,
                    modules=modules, wires=wires, instrument=instrument.id)


def make_modules(actions, prefix=""):
    """
    Convert a list of action functions into modules using auto_module.

    All ids are prefixed with prefix.
    """
    modules = []
    for action in actions:
        # Read the module defintion from the docstring
        module_description = auto_module(action)

        # Tag module ids with prefix
        module_description['id'] = prefix + module_description['id']

        # Tag each terminal data type with the data type prefix, if it is
        # not already a fully qualified name
        for v in module_description['inputs'] + module_description['outputs']:
            if '.' not in v['datatype']:
                v['datatype'] = prefix + v['datatype']

        #action = outputs_wrapper(module_description, action)
        # Define and register the module
        module = Module(action=action, **module_description)
        modules.append(module)

        #from pprint import pprint
        #pprint(module_description)

    return modules



def auto_module(action):
    """
    Given an action function, parse the docstring and return a node.

    The action name and docstring are highly stylized.

    The description is first.  It can span multiple lines.

    The parameter sections are ``**Inputs**`` and ``**Returns**``.
    These must occur on a line by themselves, with the ``**...**`` markup
    to make them show up bold in the sphinx docs.  The Inputs are split
    into input terminals and input fields depending on whether a default
    value is given in the function definition (see example below).

    Each parameter has name, type, description and optional default value.
    The name is the first on the line, followed by the type in
    (parentheses), then ':' and a description string. The parameter
    definition can span multiple lines, but the description will be
    joined to a single line.  The default value for the parameter is taken
    from the keyword argument list.  If for some reason this doesn't work,
    then add [default] in square brackets after the description, and it
    will override the keyword value.

    If the field is optional, then mark the type with '?'.  If the node
    accepts zero or more values, then mark the type with '*'.  If the node
    accepts one or more values, then mark the type with '+'.  These flags
    set the *required* and *multiple* properties of the field.

    Datasets move through the template as bundles of files. If the input
    terminals for a node are all multiple inputs, then the action will
    be called once with the entire bundle as a list.  If the input terminals
    are all single input, then the action will be called once for each dataset
    in the bundle.  For mixed single/multiple inputs the first input must
    be single.  If the output is single, it is made into a length one bundle.
    If the output is multiple, then it is assumed to already be a bundle.
    With mixed single input/multiple output, all outputs are concatenated
    into one bundle.

    All fields should be the length of the first input, or length 1 if the
    the field value is the same across all inputs, or length 0 if the
    field is optional and not defined.

    The possible types are determined by the user interface.  They include
    items such as:

        str, int, bool, float, float:units, opt:opt1|opt2|...|optn

    See :func:`parse_datatype` for the complete list.

    Each instrument will have its own data types as well, which are
    associated with the wires and have enough information that they can
    be plotted.

    Every module should have a date stamp and author, with date as
    yyyy-mm-dd so that it sorts correctly.  A change notice can be
    included, separated from author by ':'.  If there are multiple
    updates to the module, precede each with '| ' so that line breaks
    are preserved in the formatted documentation.

    The resulting doc string should look okay in sphinx. For example,

    ::

        def rescale(data, scale=1.0, dscale=0.0):
            r\"""
            Rescale the count rate by some scale and uncertainty.

            **Inputs**

            data (refldata) : data to scale

            scale (float:) : amount to scale

            dscale (float:) : scale uncertainty for gaussian error propagation

            **Returns**

            output (refldata) : scaled data

            | 2015-12-17 Paul Kienzle: first release
            | 2016-02-04 Paul Kienzle: change parameter name
            \"""

    The 'float:' type indicates that it is a unitless floating point value
    rather than a floating point value for which we forgot to specify units.

    The 'r' preceding the docstring allows us to put backslashes in the
    documentation, which is convenient when we have latex markup in between
    dollar signs or in a \\:math\\: environment.
    """
    try:
        return _parse_function(action)
    except ValueError as exc:
        annotate_exception(" while initializing module " + action.__name__, exc)
        raise



timestamp = re.compile(r"^([|] )?(?P<date>[0-9]{4}-[0-9]{2}-[0-9]{2})\s+(?P<author>.*?)\s*(:\s*(?P<change>.*?)\s*)?$")
def _parse_function(action):
    # grab arguments and defaults from the function definition
    argspec = inspect.getfullargspec(action)
    args = argspec.args if not inspect.ismethod(action) else argspec.args[1:]
    defaults = (dict(zip(args[-len(argspec.defaults):], argspec.defaults))
                if argspec.defaults else {})
    if argspec.varargs is not None or argspec.varkw is not None:
        raise ValueError("function contains *args or **kwargs")

    # Grab the docstring from the function
    # Note: inspect.getdoc() cleans the docstring, removing the indentation and
    # the leading and trailing carriage returns
    docstr = inspect.getdoc(action)

    # Default values for docstring portions
    description_lines = []
    input_lines = []
    output_lines = []
    version, author = "", ""
    changelog = []

    # Split docstring into sections
    state = 0 # processing description
    for line in docstr.split('\n'):
        match = timestamp.match(line)
        stripped = line.strip()
        if match is not None:
            state = 3
            version = match.group('date')
            author = match.group('author')
            changelog.append((version, author, match.group('change')))
        elif stripped == "**Inputs**":
            state = 1
        elif line.strip() == "**Returns**":
            state = 2
        elif state == 0:
            description_lines.append(line)
        elif state == 1:
            input_lines.append(line)
        elif state == 2:
            output_lines.append(line)
        elif state == 3:
            raise ValueError("docstring continues after time stamp")
        else:
            raise RuntimeError("Unknown state %s"%state)

    # parse the sections
    name = _unsplit_name(action.__name__)
    heading = "\n".join(("="*len(name), name, "="*len(name), ""))
    #description = "".join(description_lines)
    description = rst2html(heading + docstr, part="whole", math_output="mathjax")
    inputs = parse_parameters(input_lines)
    output_terminals = parse_parameters(output_lines)

    # Check that all defined arguments are described
    defined = set(args)
    described = set(p['id'] for p in inputs)
    if defined-described:
        raise ValueError("Parameters defined but not described: "
                         + ",".join(sorted(defined-described)))
    if described-defined:
        raise ValueError("Parameters described but not defined: "
                         + ",".join(sorted(described-defined)))

    # Make sure there are no duplicates
    all_described = set(p['id'] for p in inputs+output_terminals)
    if len(all_described) != len(inputs)+len(output_terminals):
        raise ValueError("Parameter and return value names must be unique")

    # Split parameters into terminals (non-keyword) and fields (keyword)
    field_names = set(args[-len(defaults):] if defaults else [])
    input_terminals = [p for p in inputs if p['id'] not in field_names]
    input_fields = [p for p in inputs if p['id'] in field_names]

    # Set the defaults for the fields from the keyword arguments, and make
    # sure they are consistent with the datatype.
    for p in input_fields:
        if p['datatype'] not in FIELD_TYPES:
            raise ValueError("Invalid type %s for field %s"%(p['datatype'], p['id']))
        value = defaults[p['id']]
        if value is not None:
            # Note: validate may change the value to conform to the type
            value = validate(p, value, as_default=True)
        if p['multiple']:
            value = [value] if value is not None else []
        p['default'] = value

    for p in input_terminals + output_terminals:
        if p['datatype'] in FIELD_TYPES:
            raise ValueError("Invalid type %s for terminal %s"%(p['datatype'], p['id']))

    # Collect all the node info
    result = {
        'id': action.__name__,
        'name': name,
        'description': description,
        'inputs': input_terminals,
        'outputs': output_terminals,
        'fields': input_fields,
        'version': version,
        'author': author,
        #'changelog': changelog,
        'action_id': action.__module__ + "." + action.__name__
        }

    return result


def _unsplit_name(name):
    """
    Convert "this_name" into "This Name".
    """
    return " ".join(s.capitalize() for s in name.split('_'))


# parameter definition regular expression
parameter_re = re.compile(r"""\A
    \s*(?P<id>\w+)                           # name
    \s*(\{\s*(?P<label>.*?)\s*\})?           # { label }    (optional)
    \s*(\(                                   # (
        \s*(?P<datatype>.*?)                 #    datatype  (non-greedy)
        \s*(\[\s*(?P<length>[0-9]*)\s*\])?   #    [length]  (optional)
        \s*(?P<multiple>[?*+])?              #    multiple  (optional [*+?])
        \s*(:\s*(?P<typeattr>.*?))?          #    :typeattr (optional)
    \s*\))?                                  # )
    \s*:                                     # :
    \s*(?P<description>.*?)                  # description  (non-greedy)
    \s*\Z""", re.VERBOSE)
def parse_parameters(lines):
    """
    Interpret the doc strings for the parameters.

    Each parameter must use the form defined by the following syntax:

        id {label} (type[length]#:attr): description

    The *(type)* specifier is optional, defaulting to str.  The *[default]*
    value is rarely needed since it can be extracted from the function
    defintion.  Within the type specifier, the *[length]* is not required
    if only a single value is needed.  The multiplicity specifier *#* can be
    one of '?', '*', or '+' if there can be zero or one, zero or more or one
    or more values respectively.

    *lines* is the set of lines after ``**Inputs**`` and ``**Returns**``.
    Note that parameters are defined by consecutive non-blank lines separated
    by blank lines.  :func:`get_paragraphs` is used to gather all of the
    relevant lines together, skipping the blank bits.
    """
    ret = []
    for group in get_paragraphs(lines):
        s = " ".join(s.strip() for s in group)
        match = parameter_re.match(s)
        if match is None:
            raise ValueError("unable to parse parameter:\n  "+"  ".join(group))
        d = match.groupdict()
        if d['length'] is None:
            d['length'] = 1
        elif d['length'].strip() == "":
            d['length'] = 0
        else:
            try:
                d['length'] = int(d['length'])
            except ValueError:
                raise ValueError("bad length specifier:\n "+"  ".join(group))
            if d['length'] < 0:
                raise ValueError("length must positive:\n "+"  ".join(group))
        if d['multiple'] is None:
            d['required'] = True
            d['multiple'] = False
        elif d['multiple'] == '?':
            d['required'] = False
            d['multiple'] = False
        elif d['multiple'] == '*':
            d['required'] = False
            d['multiple'] = True
        elif d['multiple'] == '+':
            d['required'] = True
            d['multiple'] = True
        else:
            raise RuntimeError("Can't get here")
        if d['label'] is None:
            d['label'] = _unsplit_name(d['id'])
        parse_datatype(d)
        ret.append(d)
    return ret

def parse_datatype(par):
    r"""
    Interpret the data type into a base type and attributes.

    *str* : String input.

    *bool:label* : True/false.

        Boolean options can be represented using a check box in the user
        interface, using the given label.  If the label is not provided,
        then the name of the parameter will be used.  If *multiple=True*,
        then there should be a separate checkbox for each dataset, which
        could be added as a check menu item on the drop down associated
        with each graph line.

    *int:<min,max>* : Integer value.

        If a range is given, then the value should lie within the range.
        The type attribute is set as *typeattr={range: [min, max]}*, with
        range defaulting to *[-inf, inf]* if no range is given.  Note that
        inf is represented as 1e9 for integers.

    *float:units<min,max>* : Floating point value.

        Units should be a string, which may be the empty string if the value
        is unitless.  The type attribute is set as
        *typeattr={units: "units", range: [min, max]}*, with range defaulting
        to *[-inf, inf]* if no range is given.  Note that inf is represented
        as 1e300 for float.

    *opt:opt1|...|optn* : Choice list.

        Select from the list of options.  Options are sent as string
        values.  This could be represented as radio buttons, as a single
        selection list or as a dropdown list, depending on the number of
        options to choose from and the amount of screen real estate available.
        If *length=0* because *opt[]:...* was specified, then a multiple
        selection list should be used to return a list of strings. If
        *multiple=True*, then the options could show up on the drop down
        menu associated with each data line.

        Normally the option name matches the option label, unless the option
        is given as "...|name=label|...". If the final choice is "...", then
        the response is an open set response, with the list being the list of
        common options.  This can be implemented using a combo box, with
        users able to enter a string or choose one of the predefined options
        from the drop-down list.  The option may be missing, in which case
        a block separation should be indicated and the option never selected.

        The type attribute is set as
        *typeattr={choices: [[opt1,label1], ...], open: true/false}*

    *regex:pattern* -> *datatype=regex, pattern=pattern*

        String which must match the given regular expression.  The type
        attribute is set as *typeattr={pattern: "pattern"}*.  Note that
        matching is not anchored, so you probably want *^pattern$*.

        To check the pattern, use::

            if (RegExp(typeattr.pattern).test(inputstr)) {}  // javascript

            if re.search(typeattr["pattern"], inputstr): pass  # python

    *fileinfo* -> *datatype=fileinfo*

        A file identifier on the server with the following structure::

            {
                *path*: location on server,
                *mtime*: modification time,
                *source*: data repository (ncnr, ncnr_DOI, local...)
            }

        The user interface should present a list of files to choose from,
        and fill in the structure.

    *index* -> *datatype=index*

        This should be a zero-origin point number in the plotted data.  There
        are several use cases, but the first one is most sensible:

        - index[]\* -> *length=0, multiple=true*

            an arbitrary list of points should be selected for each
            dataset independently

        - index\* -> *length=1, multiple=true*

            a single point should be selected for each dataset independently

        - index -> *length=1, multiple=false*

            a single point should be selected for all lines simultaneously,
            and that point will be used in all datasets.

        - index[] -> *length=0, multiple=false*

            multiple points should be selected for all lines simultaneously,
            and those points will apply to all datasets.

    *range:axis* -> *datatype=range, axis=axis*

        This is a selection of a range from the graph, where *axis* is *x* or
        *y* for a range of *x* or *y* values, or *xy* for a region of the
        graph. This makes most sense with *length=1, multiple=false*.
        The type attribute is set as *typeattr={axis: "axis"}*.  The return
        value is *[min, max]* for *x* and *y*, or *[minx, maxx, miny, maxy]*
        for *xy*.

    *coordinate* -> *datatype=coordinate*

        A point *[x, y]* should be selected from the graph.
    """
    name = par['id']
    type = par["datatype"] if par["datatype"] else "str"
    attrstr = par["typeattr"] if par["typeattr"] else ""
    attr = {}

    #print "type: %r"%type
    if type == "int":
        try:
            min, max = parse_range(attrstr, limit=1e9)
            if int(min) != min or int(max) != max:
                raise Exception("use integers for int range")
        except Exception as exc:
            raise ValueError("invalid range for %s: %s"%(name, str(exc)))
        attr["range"] = [int(min), int(max)]

    elif type == "float":
        split_index = attrstr.find("<")
        if split_index >= 0:
            units, range = attrstr[:split_index], attrstr[split_index:]
        else:
            units, range = attrstr, ""
        try:
            min, max = parse_range(range, limit=1e300)
        except Exception as exc:
            raise ValueError("invalid range for %s: %s"%(name, str(exc)))

        attr["units"] = units.strip()
        attr["range"] = [min, max]

    elif type == "opt":
        if "|" not in attrstr:
            raise ValueError("options not specified for " + name)
        attrstr = attrstr.replace('\\', '')  # allow escaped backslashes for rst
        choices = [v.split('=', 2) for v in attrstr.split("|")]
        choices = [(v*2 if len(v) == 1 else v) for v in choices]
        choices = [[v[0].strip(), v[1].strip()] for v in choices]
        open = (choices[-1][0] == "...")
        if open:
            choices = choices[:-1]

        attr["choices"] = choices
        attr["open"] = open

    elif type == "regex":
        if not attrstr:
            raise ValueError("regex is empty for " + name)
        try:
            pattern = re.compile(attrstr)
        except Exception as exc:
            raise ValueError("regex error %r for %s"%(str(exc), name))
        # if the default is given, check that it matches the pattern
        default = par.get("default", None)
        if default and not pattern.search(default):
            raise ValueError("default %r does not match pattern %r for %s"
                             %(default, attrstr, name))
        attr["pattern"] = attrstr

    elif type == "range":
        if attrstr not in ("x", "y", "xy", "ellipse", "sector_centered"):
            raise ValueError("range must be one of x, y, xy, ellipse or sector_centered for " + name)
        attr["axis"] = attrstr

    elif type == "patch_metadata":
        attr["key"] = attrstr

    else: # type in ["str", "bool", "fileinfo", "index", "coordinate"]:
        if par["typeattr"] is not None:
            raise ValueError("No restrictions on type %s for parameter %s"
                             % (type, name))


    par["datatype"] = type
    par["typeattr"] = attr

FIELD_TYPES = set("str bool int float opt regex range index coordinate fileinfo scale patch_metadata".split())

def check_multiplicity(par, values, bundle_length):
    """
    Check the number of values for the parameter against bundle length,
    depending on whether the parameter is required or multiple.
    """
    name = par["name"]
    required = par["required"]
    multiple = par["multiple"]

    if required and len(values) == 0:
        raise ValueError("need a value for %s"%name)
    if not multiple and len(values) not in [0, 1, bundle_length]:
        raise ValueError("wrong number of values for %s"%name)

def validate(par, value, as_default=False):
    """
    Check the parameter value.

    If *as_default*, make sure that the value is a reasonable default.

    Returns the value, possibly converted to the correct type.

    Raise an error if the parameter value is not valid.
    """
    name = par["id"]
    n = par["length"]
    if n == 1:
        return _validate_one(par, value, as_default)
    elif not isinstance(value, (list, tuple)):
        raise ValueError("invalid value for %s, expected list"%name)
    elif n and len(value) != n:
        raise ValueError("invalid value for %s, wrong length"%name)
    else:
        return [_validate_one(par, v, as_default) for v in value]

def _validate_one(par, value, as_default):
    datatype = par["datatype"]
    name = par["id"]

    if datatype == "int":
        # accept float values instead of integers
        value = _type_check(name, value, int)
        min, max = par["typeattr"]["range"]
        if not min <= value <= max:
            raise ValueError("invalid value for %s, %s not in [%s, %s]"
                             %(name, value, min, max))
    elif datatype == "float":
        _type_check(name, value, float)
        min, max = par["typeattr"]["range"]
        if not min <= value <= max:
            raise ValueError("invalid value for %s, %s not in [%s, %s]"
                             %(name, value, min, max))
    elif datatype == "str":
        value = _type_check(name, value, str)

    elif datatype == "opt":
        value = _type_check(name, value, str)
        choices = par["typeattr"]["choices"]
        isopenset = par["typeattr"]["open"]
        # Note: choices is a list of pairs [["option", "display"], ...]
        if value not in list(zip(*choices))[0]:
            #print value, choices
            if not isopenset or as_default is True:
                raise ValueError("value %r not in choice list for %s"
                                 %(value, name))

    elif datatype == "regex":
        value = _type_check(name, value, str)
        pattern = par["typeattr"]["pattern"]
        if not re.match(pattern, value):
            raise ValueError("value %r not does not match %r for %s"
                             %(value, pattern, name))

    elif datatype == "range":
        range = par["typeattr"]["axis"]
        n = 4 if (range == "xy" or range == "ellipse") else 2
        value = _list_check(name, value, n, float)

    elif datatype == "index":
        value = _type_check(name, value, int)

    elif datatype == "coordinate":
        value = _list_check(name, value, 2, float)

    elif datatype == "bool":
        value = _type_check(name, value, bool)

    elif datatype == "fileinfo":
        value = _finfo_check(name, value)

    elif datatype == "scale":
        value = _type_check(name, value, float)

    elif datatype == "patch_metadata":
        value = _patch_check(name, value)

    else:
        raise ValueError("no %s type check for %s"%(datatype, name))

    return value

def _finfo_check(name, value):
    try:
        if (not isinstance(value, dict)
                or "path" not in value
                or "mtime" not in value
                or "source" not in value
                or (len(value) > 3 and "entries" not in value)
                or len(value) > 4):
            raise ValueError("wrong structure")
        value["path"] = _type_check(name, value["path"], str)
        value["mtime"] = _type_check(name, value["mtime"], int)
        value["source"] = _type_check(name, value["source"], str)
        if value.setdefault("entries", None) is not None:
            value["entries"] = _list_check(name, value["entries"], 0, str)

    except:
        #raise
        raise ValueError("value %r is not {source: str, path: str, mtime: int, entries: [str, ...]} for %s"
                         % (value, name))
    return value

def _patch_check(name, value):
    if (not isinstance(value, dict)
            or "op" not in value
            or "path" not in value
            or value["op"] not in ["test", "add", "replace", "copy", "move", "remove"]):
        raise ValueError("patch must be a dict, with valid op code and path")

    elif value["op"] in ["test", "add", "replace"]:
        if not "value" in value:
            raise ValueError("patch for test, add or replace must contain value field")
        else:
            value["value"] = _type_check(name, value["value"], str)

    elif value["op"] in ["copy", "move"]:
        if not "from" in value:
            raise ValueError("patch for copy or move must contain from field")
        else:
            value["from"] = _type_check(name, value["from"], str)

    value["op"] = _type_check(name, value["op"], str)
    value["path"] = _type_check(name, value["path"], str)

    return value

def _list_check(name, values, n, ptype):
    #print "list check", name, values, n, ptype
    if isinstance(values, tuple):
        values = list(values)
    _type_check(name, values, list)
    if n and len(values) != n:
        raise ValueError("expected list of length %d for %s"
                         %(n, name))
    try:
        values = [_type_check(name, v, ptype) for v in values]
    except:
        raise ValueError("all values in list must be of type %s for %s"
                         %(str(ptype), name))
    return values

def _type_check(name, value, ptype):
    if value is None:
        return value
    elif ptype is int and isinstance(value, float) and int(value) == value:
        value = int(value)
    elif ptype is float and isinstance(value, int):
        value = float(value)
    elif ptype is str and not IS_PY3 and isinstance(value, unicode):
        value = value.encode('utf-8')
    elif ptype is str and IS_PY3 and isinstance(value, bytes):
        value = value.decode('utf-8')
    if not isinstance(value, ptype):
        raise ValueError("expected %s for %s but got %s"
                         % (str(ptype), name, str(type(value))))
    return value


def parse_range(range, limit=inf):
    """
    Parse the range string *<min,max>*.

    *limit* sets the limiting value for the range, to use in place of inf or
    a missing limit.

    Returns (*min, max*), or *(-limit, limit)* if no range is provided.
    """
    range = range.strip()
    if range == "":
        return [-limit, limit]
    if not (range.startswith('<') and range.endswith('>') and "," in range):
        raise ValueError("expected <min, max>")
    minstr, maxstr = [v.strip() for v in range[1:-1].split(',', 2)]
    min = float(minstr) if minstr and minstr != "-inf" else -limit
    max = float(maxstr) if maxstr and maxstr != "inf" else limit
    if min >= max:
        raise ValueError("min must be less than max in (%g,%g)"%(min, max))
    if min < -limit:
        raise ValueError("min value %g must be more than %g"%(min, -limit))
    if max > limit:
        raise ValueError("max value %g must be more than %g"%(max, limit))
    return min, max

def get_paragraphs(lines):
    """
    Yield a list of paragraphs defined as lines separated by blank lines.

    Each paragraph is returned as a list of lines.
    """
    group = []
    for line in lines:
        if line.strip() == "":
            if group:
                yield group
            group = []
        else:
            group.append(line)
    if group:
        yield group


def test_parse_parameters():
    all_good = """
        s1 (str) : p1 description

        s2 (str) : colon not allowed after str.
        Description can go on for many lines.
        Default value is still at the end

        s3 : input defaults to string

        i1 (int?) : optional integer
            extended description can be indented.

        i2 (int*:<0,5>) : multiple integers in <0,5>

        i3 (int[]:<3,>) : integer >= 3

        i4 ( int [ ] * : < 3, inf > ) : integer >= 3

        i5(int+:<,6>):atleastoneinteger<=6

        i6 (int:<-inf,-6>) : i <= -6

        f1 (float:s) : floating point with seconds unit

        f2 (float:<2.718, 3.142>) :floating point with range and no units

        f3 (float:neutrons/s<0,inf>) : count rates must be non-negative

        f4 ( float: neutrons/s ) : count rates must be non-negative

        f5 (float: neutrons/s <0,inf> ) : count rates must be non-negative

        f6 (float[3]) : x,y,z coordinates

        o1 (opt:a|b|c) : simple options

        o2 (opt*: space in name | allowed | ... ) : spaces in options

        o3 (opt[]: long|options
            |split) : over multiple lines

        o4 (opt: p1=Parameter 1||Parameter 2) : empty option allowed

        e1 (regex:.*) :

        r1 (range:x) : x range

        r2 (range:y ) : y range

        r3 (range: xy ) : xy range

        spinflip {Correct Spinflip Data} (bool)
            : Correct spinflip data if available.

        file_name (fileinfo) :

        point (index) :

        coord (coordinate) :
    """

    p = dict((pk['id'], pk) for pk in parse_parameters(all_good.split('\n')))
    #import pprint; pprint.pprint(p)

    assert p['s1']['description'] == 'p1 description'
    assert p['i1']['description'] == 'optional integer extended description can be indented.'
    assert p['e1']['description'] == ''

    assert p['spinflip']['label'] == 'Correct Spinflip Data'
    assert p['file_name']['label'] == 'File Name'
    assert p['r3']['label'] == 'R3'

    # See FIELD_TYPES for list of available field types
    assert p['s3']['datatype'] == 'str'
    assert p['spinflip']['datatype'] == 'bool'
    assert p['i2']['datatype'] == 'int'
    assert p['f4']['datatype'] == 'float'
    assert p['o2']['datatype'] == 'opt'
    assert p['e1']['datatype'] == 'regex'
    assert p['r1']['datatype'] == 'range'
    assert p['point']['datatype'] == 'index'
    assert p['coord']['datatype'] == 'coordinate'
    assert p['file_name']['datatype'] == 'fileinfo'

    assert p['f5']['required'] == True   # multiplicity: none
    assert p['f5']['multiple'] == False
    assert p['i1']['required'] == False  # multiplicity: ?
    assert p['i1']['multiple'] == False
    assert p['i5']['required'] == True   # multiplicity: +
    assert p['i5']['multiple'] == True
    assert p['o2']['required'] == False  # mulitplicity: *
    assert p['o2']['multiple'] == True

    assert p['f4']['length'] == 1  # no vector tag
    assert p['i4']['length'] == 0  # tagged with [ ]
    assert p['f6']['length'] == 3  # tagged with [3]

    assert p['s1']['typeattr'] == {}

    assert p['i1']['typeattr']['range'] == [-1e9, 1e9]
    assert p['i2']['typeattr']['range'] == [0, 5]
    assert p['i3']['typeattr']['range'] == [3, 1e9]
    assert p['i4']['typeattr']['range'] == [3, 1e9]
    assert p['i5']['typeattr']['range'] == [-1e9, 6]
    assert p['i6']['typeattr']['range'] == [-1e9, -6]

    assert p['f1']['typeattr']['units'] == 's'
    assert p['f2']['typeattr']['units'] == ''
    assert p['f3']['typeattr']['units'] == 'neutrons/s'
    assert p['f4']['typeattr']['units'] == 'neutrons/s'
    assert p['f5']['typeattr']['units'] == 'neutrons/s'

    assert p['f1']['typeattr']['range'] == [-1e300, 1e300]
    assert p['f2']['typeattr']['range'] == [2.718, 3.142]

    assert p['o1']['typeattr']['choices'] == [
        ['a', 'a'], ['b', 'b'], ['c', 'c']
    ]
    assert p['o2']['typeattr']['choices'] == [
        ['space in name', 'space in name'], ['allowed', 'allowed']
    ]
    assert p['o3']['typeattr']['choices'] == [
        ['long', 'long'], ['options', 'options'], ['split', 'split']
    ]
    assert p['o4']['typeattr']['choices'] == [
        ['p1', 'Parameter 1'], ['', ''], ['Parameter 2', 'Parameter 2']
    ]

    assert p['e1']['typeattr'] == {'pattern': '.*'}

    assert p['r1']['datatype'] == 'range'
    assert p['r1']['typeattr'] == {'axis': 'x'}



#!/usr/bin/env python
"""
Data loader for Rigaku table-top X-ray source .ras file format.

Provides :func:`load` and :func:`loads` functions to load a Rigaku RAS format
file.  These functions return a list of file segments containing the
following::

    sample : sample name
    comment : file comment

    x : x data
    y : scaled counts, not accounting for count time or attenuators
    y_err : count uncertainty for scaled counts

    x_label, x_unit : x axis label and units
    y_label, y_unit : y axis label and units

    x_resolution : uncertainty in x position?
    wavelength : wavelength of the beam (Angstroms)
    wavelength_resolution : 1-sigma wavelength resolution
    slit1_distance : negative distance from sample to selection slit
    slit2_distance : negative distance from sample to incident slit
    slit3_distance : distance from sample to receiving slit 1
    slit4_distance : distance from sample to receiving slit 2

    scan_mode : STEP | ?
    scan_axis : which motors are being scanned
        TwoThetaOmega, TwoThetaTheta
            sample angle = x/2 + Omega - TwoTheta/2
            detector angle = x
        Omega
            sample angle = x
            detector angle = TwoTheta
        TwoTheta
            sample angle = Omega
            detector angle = x

    count_time, count_time_unit : counting time and units for the segment
    start_time, end_time : start and end time for segment, as struct_time

    axis : { name: [label, unit, position, offset] } instrument configuration,
    including the following fields

        TwoTheta, Omega
        Chi, Phi
        Z, Rx, Ry
        TwoThetaTheta, TwoThetaOmega, TwoThetaChi, TwoThetaChiPhi
        Alpha, Beta
        ThetaS, ThetaD
        Ts, Zs
        CBO, CBO-M
        Incident{SollerSlit, AxdSlit, SlitBox{,-_Axis}, Monochromator{,-OmegaM}}
        CenterSlit, Filter, Attenuator
        Receiving{SollerSlit, AxdSlit, SlitBox[12]{,-_Axis}, Optics, SlitBox2-Zd}
        Counter{Monochromator, Slit}
        TwoThetaB, AlphaR, BetaR
        IncidintPrimary, HV, PHA

    full_header : all the Rigaku header fields as string float or int values

The CBO axis (Cross Beam Optics) can be 'BB' for Bragg-Brentano divergent
beam optics or 'PB' for parallel beam optics.  Convergent beam optics
requires an elliptical mirror rather than the parabolic mirror used for the
parallel beam optics. Likely this will be indicated by the presence
of the 'CBO-E unit' as one of the 'HW_I_OPT_NAME-#' fields.  Perhaps the CBO
axis will use the code 'CB' instead of 'PB' for this case?

Slit distances are not stored in the file, and instead came from the schematic
from the Rigaku SmartLab instrument.

Wavelength is determined by target and monochromator, with the file containing
values for K_a1, K_a2 and K_b.  With the Ge(220)x2 monochromator, it is possible
to select just 'Ka1', as indicated by the value in 'MEAS_COND_XG_WAVE_TYPE'.
K_alpha linewidths are tiny[1], mostly less than 0.05% FWHM.  This value is
used to set the wavelength error independent of target 'HW_XG_TARGET_NAME'.
For high intensity measurements use (2*Ka1 + Ka2)/3 for the wavelength, and
standard deviation of (Ka1, Ka1, Ka2) for the 1-sigma dispersion.  Note: do not
yet know how this is indicated in the file.

[1] https://wwwastro.msfc.nasa.gov/xraycal/linewidths.html
"""

from __future__ import division, print_function
import warnings
import sys
import traceback
import logging
from time import strptime

import numpy as np

if sys.version_info[0] >= 3:
    def tostr(s):
        return s.decode('ascii')
else:
    def tostr(s):
        return s

# Time format in Rigaku .ras files is "mm/dd/yy HH:MM:SS"
TIME_FORMAT = "%m/%d/%y %H:%M:%S"
NEW_TIME_FORMAT = "%m/%d/%Y %H:%M:%S"
FWHM = np.sqrt(np.log(256))

def parse_time(timestr):
    try:
        return strptime(timestr, TIME_FORMAT)
    except ValueError:
        return strptime(timestr, NEW_TIME_FORMAT)

# Copied from anno_exc so that rigaku.py is stand-alone
def annotate_exception(msg, exc=None):
    """
    Add an annotation to the current exception, which can then be forwarded
    to the caller using a bare "raise" statement to reraise the annotated
    exception.
    """
    if not exc:
        exc = sys.exc_info()[1]

    args = exc.args
    if not args:
        arg0 = msg
    else:
        arg0 = " ".join((str(args[0]), msg))
    exc.args = tuple([arg0] + list(args[1:]))

def load(filename):
    """
    Load Rigaku data from file.
    """
    with open(filename, 'rb') as fid:
        data = fid.read()
    try:
        return loads(data)
    except Exception as exc:
        #traceback.print_exc()
        annotate_exception("while loading %r"%filename)
        raise

def loads(data):
    """
    Load Rigaku data from string.
    """
    lines = data.split(b'\r\n')
    sections = _split_sections(lines)

    # skip sections that are not 'HEADER" or "INT"
    sections = [(name, body, index) for name, body, index in sections
                if name in set((b'HEADER', b'INT'))]

    for name, body, index in sections[::2]:
        if name != b'HEADER':
            # Note: index-1 for *RAS_section_START and index+1 for 1-origin
            raise ValueError("Expected *RAS_HEADER_START at %d"%index)
    for name, body, index in sections[1::2]:
        if name != b'INT':
            # Note: index-1 for *RAS_section_START and index+1 for 1-origin
            raise ValueError("Expected *RAS_INT_START at %d"%index)

    head = [_parse_header(body, index+1) for name, body, index in sections[::2]]
    vals = [_parse_data(body, index+1) for name, body, index in sections[1::2]]
    if len(head) != len(vals):
        raise ValueError("Missing data block for final section")
    datasets = [_interpret(h, v) for h, v in zip(head, vals)]

    return datasets

def join(datasets):
    """
    Combine Rigaku data segments into a single scan.
    """
    first = datasets[0]
    if any(data['scan_axis'] != first['scan_axis']
           for data in datasets):
        raise ValueError("Can't mix different scan types")
    if any(data['axis']['Attenuator'][2] != first['axis']['Attenuator'][2]
           for data in datasets):
        raise ValueError("Can't mix different attenuation values")

    R = first.copy()
    R['x'] = np.hstack([data['x'] for data in datasets])
    R['y'] = np.hstack([data['y'] for data in datasets])
    R['y_err'] = np.hstack([data['y_err'] for data in datasets])
    R['count_time'] = np.hstack([data['count_time']*np.ones_like(data['x'])
                                 for data in datasets])
    R['start_time'] = min(data['start_time'] for data in datasets)
    R['end_time'] = max(data['end_time'] for data in datasets)
    return R

def _split_sections(lines):
    """
    Split rigaku file into sections

    Returns a list of sections [(name, body, index)] where name is the section
    type, body is the list of lines between *RAS_name_START and *RAS_name_END
    and index is the 1-origin index of *RAS_name_START (or the 0-origin index
    of the first line of the body).
    """
    if lines[0] != b"*RAS_DATA_START":
        raise ValueError("not a Rigaku XRD RAS file")

    last_index = len(lines) - 1
    while last_index > 0 and lines[last_index] != b"*RAS_DATA_END":
        if lines[last_index] != b"":
           warnings.warn("Non-empty line found after *RAS_DATA_END: '%s' at line %d"%(lines[last_index], last_index))
        last_index -= 1

    sections = []
    end_target = None
    start_index = 1000000000 # initialize with int to keep linter happy
    for index, line in enumerate(lines[1:last_index]):
        if line == end_target:
            # Note: index chosen so that the section includes
            # "*RAS_section_START" but not "*RAS_section_END"
            name = lines[start_index+1][5:-6]
            body = lines[start_index+2:index+1]
            sections.append((name, body, start_index+1))
            end_target = None
        elif end_target is None:
            if not (line.startswith(b"*RAS_") and line.endswith(b"_START")):
                raise ValueError("Rigaku expected section start at line %d"
                                 %(index+2))
            start_index = index
            end_target = line[:-6] + b"_END"
        #else: accumulate into current section
    if end_target is not None:
        raise ValueError("Rigaku file incomplete: final section is not closed")
    return sections

def _parse_header(lines, start_index):
    """
    Interpret the key-value pairs in the HEADER section.
    """
    header = {}
    for offset, line in enumerate(lines):
        try:
            key, value = line.split(b' ', 1)  # *KEY "value"
            # Note: py3 byte strings key[k] returns ord not byte string
            assert key[0:1] == b"*"
            assert value[0:1] == b'"' and value[-1:] == b'"'
        except Exception:
            #traceback.print_exc()
            raise ValueError("corrupt file: line %d is not '*KEY value'"
                             %(start_index + offset))
        key = tostr(key[1:])
        value = value[1:-1] # strip quotes

        # auto convert values to int or float if possible
        # for string values, try splitting "japanese?|english" to english
        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                try:
                    _, value = value.split(b'|')
                except ValueError:
                    pass
                value = tostr(value)
        # if all conversions fail, value should be an untouched string

        header[key] = value
    return header

def _parse_data(lines, start_index):
    """
    Interpret the list of column values in the INT section.
    """
    values = []
    for offset, line in enumerate(lines):
        try:
            values.append([float(v) for v in line.split()])
        except Exception:
            raise ValueError("corrupt file: line %d is not a set of values"
                             %(start_index + offset))
    return values


# From https://wwwastro.msfc.nasa.gov/xraycal/linewidths.html
EMISSION_LINEWIDTH = 0.00021  # 0.05% dL/L FWHM as 1-sigma line width

# From Rigaku manual, pg 17, pg 131
MONOCHROMATOR_WAVELENGTH_RESOLUTION = {
    # 'mirror': Ka1 + Ka2 + Kb
    'Ge(220)x2': 3.8e-4/FWHM,
    'Ge(400)x2': np.nan,  # Manual doesn't list the resolution for this config.
    'Ge(220)x4': 1.5e-4/FWHM,
    'Ge(440)x4': 2.3e-5/FWHM,
}

# From Rigaku manual, pg 131, converted from seconds of arc to degrees
MONOCHROMATOR_ANGULAR_DIVERGENCE = {
    'mirror': 4.1e-2/FWHM,
    'Ge(220)x2': 8.8e-3/FWHM,
    'Ge(400)x2': 1.2e-2/FWHM,
    'Ge(220)x4': 3.4e-3/FWHM,
    'Ge(440)x4': 1.5e-3/FWHM,
}
DIRECT_ANGULAR_DIVERGENCE = 150./3600./FWHM
# Note: receiving resolution (pg 17) is a factor of three lower than expected
# from the equation 1/2 (s1 + s2) / |d1 - d2|.  If the manual is giving values
# as 1-sigma, then they are 25% lower than expected.

def _calc_count_time(header):
    """
    Determine the count time from header constants

    returns value in seconds
    """
    scan_speed = header['MEAS_SCAN_SPEED']
    scan_speed_unit = (header['MEAS_SCAN_SPEED_UNIT']).lower()
    if scan_speed_unit == "sec.":
        return scan_speed
    elif scan_speed_unit.endswith("/min"):
        if np.isclose(scan_speed, 0.0):
            raise ValueError("scan speed can not be zero in scanning mode")
        scan_step = header['MEAS_SCAN_STEP']
        # then count_time (sec) = scan_step (step_unit) * 60 (sec/min) * 1/scan_speed (min/step_unit)
        return scan_step * 60. / scan_speed
    else:
        raise ValueError("scan speed unit \"%s\" is not recognized (should be 'sec.' or end with '/min')" % scan_speed_unit)

def _interpret(header, values):
    R = {}
    x, I, scale = np.array(values).T
    R['x'] = x
    R['y'] = I*scale
    R['y_err'] = np.sqrt(I)*scale

    R['x_label'] = header['MEAS_SCAN_AXIS_X']
    R['x_unit'] = header['MEAS_SCAN_UNIT_X']
    R['x_resolution'] = header['MEAS_SCAN_RESOLUTION_X']
    R['y_unit'] = header['MEAS_SCAN_UNIT_Y']
    R['y_label'] = header.get('DISP_TITLE_Y', 'y')

    R['count_time'] = _calc_count_time(header)
    R['count_time_unit'] = 'seconds'

    R['sample'] = header['FILE_SAMPLE']
    R['comment'] = header['FILE_COMMENT']
    R['start_time'] = parse_time(header['MEAS_SCAN_START_TIME'])
    R['end_time'] = parse_time(header['MEAS_SCAN_END_TIME'])
    R['axis'] = _interpret_axes(header)
    R['scan_axis'] = header['MEAS_SCAN_AXIS_X_INTERNAL']
    R['scan_mode'] = header['MEAS_SCAN_MODE']
    #R['scan_steps'] = (header['MEAS_SCAN_START'], header['MEAS_SCAN_STEP'],
    #                   header['MEAS_SCAN_STOP'])

    R['slit1_distance'] = 114.-300.  # mm  Selection slit
    R['slit2_distance'] = 190.-300.  # mm  Incident slit
    R['slit3_distance'] = 187.  # mm  Receiving slit 1
    R['slit4_distance'] = 300.  # mm  Receiving slit 2

    monochromator = R['axis']['IncidentMonochromator'][2]
    #print("monochromator", monochromator)
    R['wavelength_resolution'] \
        = MONOCHROMATOR_WAVELENGTH_RESOLUTION.get(monochromator, np.nan)
    R['angular_divergence'] \
        = MONOCHROMATOR_ANGULAR_DIVERGENCE.get(monochromator, DIRECT_ANGULAR_DIVERGENCE)
    if header['MEAS_COND_XG_WAVE_TYPE'] == "Ka1":
        R['wavelength'] = header['HW_XG_WAVE_LENGTH_ALPHA1']
    elif header['MEAS_COND_XG_WAVE_TYPE'] == "Ka2":
        R['wavelength'] = header['HW_XG_WAVE_LENGTH_ALPHA1']
    elif header['MEAS_COND_XG_WAVE_TYPE'] == "Kb":
        R['wavelength'] = header['HW_XG_WAVE_LENGTH_BETA']
    else:
        # TODO: missing K_beta contribution
        Ka1 = header['HW_XG_WAVE_LENGTH_ALPHA1']
        Ka2 = header['HW_XG_WAVE_LENGTH_ALPHA2']
        Kbeta = header['HW_XG_WAVE_LENGTH_BETA']
        R['wavelength'] = (2*Ka1 + Ka2)/3
        R['wavelength_resolution'] = np.std([Ka1, Ka1, Ka2], ddof=1)

    R['full_header'] = header
    return R

def _interpret_axes(header):
    # set a default attenuator, in case one is not defined in header:
    axis = {
        "Attenuator": ("Attenuator", "", 1.0, 0.0)
    }
    idx = 0
    while 'MEAS_COND_AXIS_NAME-%d'%idx in header:
        # Axis properties as string values direct from the header
        label = header['MEAS_COND_AXIS_NAME-%d'%idx]
        unit = header['MEAS_COND_AXIS_UNIT-%d'%idx]
        name = header['MEAS_COND_AXIS_NAME_INTERNAL-%d'%idx]
        #magicno = header['MEAS_COND_AXIS_NAME_MAGICNO-%d'%idx]
        offset = header['MEAS_COND_AXIS_OFFSET-%d'%idx]
        position = header.get('MEAS_COND_AXIS_POSITION-%d'%idx, float('nan'))

        # Convert position to float if possible.  Note that there are
        # non-float values for positions, such as "1/10000" for attenuators
        # and "0.2mm" for slit boxes.  There are also string names
        # such "PSA_open" for the receiving optical device field.
        if name == "Attenuator":
            # Attenuator position may be represented as a fraction, such as 1/10000.
            # Not how attenuator position is represented if absent, so I'm
            # testing a bunch of different possible strings.  Or maybe it is a
            # numeric value such as "1".
            if "/" in position:
                numerator, denominator = position.split('/')
                position = float(numerator)/float(denominator)
            elif position in ("", "-", "None", "Open"):
                position = 1.0
            # Maybe test for zero, or maybe not.  It is possible that the
            # attenuation factor is set to zero when there is a beam stop
            # in place for dark beam measurements.  Leave alone for now
            # until we have a dataset which has attenuator==0.
            #if position == 0.: position = 1.
        elif not isinstance(position, (int, float)) and position.endswith('mm'):
            unit = 'mm'
            position = float(position[:-2])
        elif position == "None":
            position = None
        elif position == "-":
            position = np.nan

        # Offsets are much easier: either "-" or a floating point number
        if offset == "-":
            offset = np.nan

        axis[name] = label, unit, position, offset
        idx += 1
    return axis

def plot(datasets, joined=True, label=None):
    from matplotlib import pyplot as plt
    if joined:
        R = join(datasets)
        plt.errorbar(R['x'], R['y']/R['count_time'], yerr=R['y_err']/R['count_time'],
                     fmt='.', label=label)
        plt.xlabel("%s (%s)" % (R['x_label'], R['x_unit']))
        plt.ylabel("%s (%s/%s)" % (R['y_label'], R['y_unit'], R['count_time_unit']))
    else:
        for k, data in enumerate(datasets):
            attenuator = data['axis']['Attenuator'][2]
            count_time = data['count_time']
            scale = 1/(attenuator*count_time)
            plt.errorbar(data['x'], data['y']*scale, yerr=data['y_err']*scale,
                         fmt='.', label=(label + "-" + str(k) if label else label))
        plt.xlabel("%s (%s)" % (data['x_label'], data['x_unit']))
        plt.ylabel("%s (%s/%s)" % (data['y_label'], data['y_unit'], data['count_time_unit']))
    plt.yscale('log')
    plt.legend(loc='best')

def show_header_diff(datasets):
    keys = set.union(*(set(data['full_header'].keys()) for data in datasets))
    values = {k: [] for k in keys}
    for data in datasets:
        for k in keys:
            values[k].append(data['full_header'].get(k, None))
    for k, v in sorted(values.items()):
        if any(vk != v[0] for vk in v):
            print(k, v)

def main_pprint():
    from pprint import pprint
    pprint(load(sys.argv[1]))

def main_plot():
    from matplotlib import pyplot as plt
    for filename in sys.argv[1:]:
        datasets = load(filename)
        plot(datasets, joined=len(sys.argv) > 2, label=filename)
    plt.show()

def main_headers():
    from pprint import pprint
    show_header_table(load(sys.argv[1]))

if __name__ == "__main__":
    #main_pprint()
    main_plot()
    #main_headers()



#!/usr/bin/env python
from warnings import warn
from copy import copy

import numpy as np

from reductus.dataflow.lib.exporters import exports_json

from .refldata import ReflData, Intent, Group, Detector, set_fields
from .nexusref import load_nexus_entries, nexus_common, get_pol
from .nexusref import data_as, str_data
from .nexusref import TRAJECTORY_INTENTS
from .resolution import FWHM2sigma

def load_metadata(filename, file_obj=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    return load_nexus_entries(filename, file_obj=file_obj,
                              meta_only=True, entry_loader=Candor)

def load_entries(filename, file_obj=None, entries=None):
    #print("loading", filename, file_obj)
    return load_nexus_entries(filename, file_obj=file_obj, entries=entries,
                              meta_only=False, entry_loader=Candor)

#: Number of detector channels per detector tube on CANDOR
NUM_CHANNELS = 54

# CRUFT: these will be in the nexus/config.js eventually
S1_DISTANCE = -4335.86
S2_DISTANCE = -356.0
S3_DISTANCE = 356.0
DETECTOR_DISTANCE = 3496.0
MONO_WAVELENGTH = 4.75
MONO_WAVELENGTH_DISPERSION = 0.01

# CRUFT: these will be in the datafile eventually
_EFFICIENCY = None
def detector_efficiency():
    from pathlib import Path
    global _EFFICIENCY
    if _EFFICIENCY is None:
        path = Path(__file__).parent / 'DetectorWavelengths.csv'
        eff = np.loadtxt(path, delimiter=',')[:, 3]
        eff = eff / np.mean(eff)
        _EFFICIENCY = np.vstack((eff, eff)).T
    return _EFFICIENCY

@set_fields
class Attenuator(Group):
    """
    Define built-in attenuators
    This is used by the attenuation correction.
    
    The wavelength-dependent transmission is defined by a fitted
    polynomial for each attenuator, along with a 
    covariance matrix for uncertainty propagation.

    transmission (n x m)
        transmission vs. wavelength
        length n is the number of defined attenuators, m is number of detectors
    transmission_err (n x m)
        1-sigma width of uncertainty distribution for transmission matrix
    target_value (npts)
        attenuator setting, for each data point (in [0, n])
    """
    transmission = None # n x m matrices
    transmission_err = None # m x m matrices
    target_value = None # attenuators in the beam; setting is per point, matching length of counts

class Candor(ReflData):
    """
    Candor data entry.

    See :class:`refldata.ReflData` for details.
    """
    format = "NeXus"
    probe = "neutrons"
    _groups = ReflData._groups + (("attenuator", Attenuator),)
    attenuator = None

    def __init__(self, entry, entryname, filename):
        super().__init__()
        self.attenuator = Attenuator()
        nexus_common(self, entry, entryname, filename)
        self.geometry = 'vertical'
        self.align_intensity = "slit1.x"

    def load(self, entry):
        #print(entry['instrument'].values())
        das = entry['DAS_logs']
        n = self.points
        raw_intent = str_data(das, 'trajectoryData/_scanType')
        if raw_intent in TRAJECTORY_INTENTS:
            self.intent = TRAJECTORY_INTENTS[raw_intent]

        # Figure out beam mode:
        #    Q.beamMode = WHITE_BEAM|SINGLE_BEAM
        #    convergingGuideMap.key = IN|OUT
        # For white beam with non-converging guide use the active bank to
        # determine which Q values to look at. For now look at both banks.
        #beam_mode = str_data(das, 'Q/beamMode', '')
        #if beam_mode != 'WHITE_BEAM':
        #    raise ValueError(f"candor reduction requires Q/beamMode=WHITE_BEAM in '{self.path}'")
        #active_bank = data_as(das, 'Q/angleIndex', 1) - 1
        #active_channel = data_as(das, 'Q/wavelenghtIndex', 1) - 1
        #is_converging = (str_data(das, 'convergingGuidesMap/key', 'OUT') == 'IN')

        # Polarizers
        self.polarization = (
            get_pol(das, 'frontPolarization')
            + get_pol(das, 'backPolarization')
        )

        # Counts
        # Load counts early so we can tell whether channels are axis 1 or 2
        counts = data_as(das, 'multiDetector/counts', '', dtype='d')
        if counts is None: # CRUFT: NICE Ticket #00113618 - Renamed detector from area to multi
            counts = data_as(das, 'areaDetector/counts', '', dtype='d')
        if counts is None or counts.size == 0:
            raise ValueError("Candor file '{self.path}' has no area detector data.".format(self=self))

        channels_at_end = (counts.shape[2] == NUM_CHANNELS)
        if channels_at_end:
            counts = np.swapaxes(counts, 1, 2)
        self.detector.counts = counts
        self.detector.counts_variance = counts.copy()
        self.detector.dims = counts.shape[1:]

        # Monochromator
        # Note: could test Q/beamMode == 'SINGLE_BEAM" for is_mono
        is_mono = (str_data(das, 'monoTrans/key', 'OUT') == 'IN')
        if is_mono:
            self.warn("monochromatic beams not yet supported for Candor reduction")
            self.monochromator.wavelength = data_as(entry, 'instrument/monochromator/wavelength', 'Ang', rep=n)
            # TODO: make sure wavelength_error is 1-sigma, not FWHM %
            self.monochromator.wavelength_resolution = data_as(entry, 'instrument/monochromator/wavelength_error', '', rep=n)
            # CRUFT: nexus config doesn't link in monochromator
            if self.monochromator.wavelength is None:
                self.monochromator.wavelength = data_as(das, 'mono/wavelength', 'Ang', rep=n)
            if self.monochromator.wavelength is None:
                self.warn("Wavelength is missing; using {MONO_WAVELENGTH} A".format(MONO_WAVELENGTH=MONO_WAVELENGTH))
                self.monochromator.wavelength = MONO_WAVELENGTH
            # TODO: generate wavelength_error directly with the nexus writer
            if self.monochromator.wavelength_resolution is None:
                dispersion = data_as(das, 'mono/wavelengthSpread', '', rep=n)
                if dispersion is not None:
                    self.monochromator.wavelength_resolution = \
                        FWHM2sigma(dispersion*self.monochromator.wavelength)
            if self.monochromator.wavelength_resolution is None:
                self.warn("Wavelength resolution is missing; using %.1f%% dL/L FWHM"
                        % (100*MONO_WAVELENGTH_DISPERSION,))
                self.monochromator.wavelength_resolution = \
                    FWHM2sigma(MONO_WAVELENGTH_DISPERSION*self.monochromator.wavelength)
            self.monochromator.wavelength = self.monochromator.wavelength[:, None, None]
            self.monochromator.wavelength_resolution = self.monochromator.wavelength_resolution[:, None, None]
        else:
            self.monochromator.wavelength = None
            self.monochromator.wavelength_resolution = None

        # Slit distances
        self.slit1.distance = data_as(entry, 'instrument/presample_slit1/distance', 'mm')
        self.slit2.distance = data_as(entry, 'instrument/presample_slit2/distance', 'mm')
        self.slit3.distance = data_as(entry, 'instrument/predetector_slit1/distance', 'mm')
        self.slit4.distance = data_as(entry, 'instrument/predetector_slit2/distance', 'mm')
        # CRUFT: old candor files don't populate instrument slits
        for slit, default in (
                (self.slit1, S1_DISTANCE),
                (self.slit2, S2_DISTANCE),
                (self.slit3, S3_DISTANCE),
                (self.slit4, DETECTOR_DISTANCE),
                ):
            if slit.distance is None:
                slit.distance = default

        # Slit openings
        # Note: Candor does not have slit4, but we translate the detector
        # mask value into a slit4 opening positioned right at the detector.
        # TODO: get values from instrument group instead of DAS_logs.
        for k, slit in enumerate((self.slit1, self.slit2, self.slit3, self.slit4)):
            x = 'slitAperture%d/softPosition'%(k+1)
            x_target = 'slitAperture%d/desiredSoftPosition'%(k+1)
            slit.x = data_as(das, x, 'mm', rep=n)
            slit.x_target = data_as(das, x_target, 'mm', rep=n)
        # There is one vertical slit, and it is defined by slit2AVertical
        # This is a virtual device which gives the center and opening, but
        # not the target center and opening. Instead, use the Top/Bottom motors
        # to get the desired and actual, and ignore slit2AVertical. This assumes
        # the slits are centered.
        self.slit2.y = (
            data_as(das, 'slit2AVerticalTop/softPosition', 'mm', rep=n)
            - data_as(das, 'slit2AVerticalBottom/softPosition', 'mm', rep=n))
        self.slit2.y_target = (
            data_as(das, 'slit2AVerticalTop/desiredSoftPosition', 'mm', rep=n)
            - data_as(das, 'slit2AVerticalBottom/desiredSoftPosition', 'mm', rep=n))
        # CRUFT: old files don't have the detector mask mapped
        # TODO: don't throw away datasets where the key is not defined:
        #       instead figure out an effective aperture size from motors or
        #       allow it to be overridden
        if self.slit4.x is None:
            mask_str = str_data(das, 'detectorMaskMap/key', '10')
            if not mask_str:
                mask_str = '10'
            try:
                mask = float(mask_str)
            except ValueError:
                raise ValueError("mask value {mask_str} cannot be converted to float".format(mask_str=mask_str))
            self.slit4.x_target = self.slit4.x = np.full(n, mask)

        # Detector
        wavelength = data_as(das, 'detectorTable/wavelengths', '')
        wavelength_spread = data_as(das, 'detectorTable/wavelengthSpreads', '')
        divergence = data_as(das, 'detectorTable/angularSpreads', '')
        efficiency = data_as(das, 'detectorTable/detectorEfficiencies', '')
        if divergence is None:
            divergence = np.repeat(0.01, wavelength.shape)
        if efficiency is None:
            efficiency = np.repeat(1.0, wavelength.shape)

        # PAK 2020-11-12: detectorTable now uses [54,2] rather than [108], so
        # reshape is no longer necessary. Check for [2,54] and transpose to
        # protect against future change. For older data, use reshape with
        # wavelengths in "channels_at_end" order (even though the detector
        # counts are in "banks_at_end" order).
        if all(dim != NUM_CHANNELS for dim in wavelength.shape):
            wavelength = wavelength.reshape(NUM_CHANNELS, -1)
            wavelength_spread = wavelength_spread.reshape(NUM_CHANNELS, -1)
            divergence = divergence.reshape(NUM_CHANNELS, -1)
            efficiency = efficiency.reshape(NUM_CHANNELS, -1)
        elif wavelength.shape[0] != NUM_CHANNELS:
            wavelength = wavelength.T
            wavelength_spread = wavelength_spread.T
            divergence = divergence.T
            efficiency = efficiency.T

        wavelength_resolution = wavelength_spread # FWHM2sigma(wavelength * wavelength_spread)
        if (efficiency == 1.0).all():
            efficiency = detector_efficiency()

        self.detector.wavelength = wavelength[None, :, :]
        self.detector.wavelength_resolution = wavelength_resolution[None, :, :]
        self.detector.efficiency = efficiency[None, :, :]
        #self.angular_resolution = divergence[None, :, :]
        # TODO: sample broadening?

        # Angles
        self.sample.angle_x = data_as(das, 'sampleAngle/softPosition', 'degree', rep=n)
        self.sample.angle_x_target = data_as(das, 'sampleAngle/desiredSoftPosition', 'degree', rep=n)
        # CRUFT: renamed sampleAngleMotor to sampleAngle
        if self.sample.angle_x is None:
            self.sample.angle_x = data_as(das, 'sampleAngleMotor/softPosition', 'degree', rep=n)
        if self.sample.angle_x_target is None:
            self.sample.angle_x_target = data_as(das, 'sampleAngleMotor/desiredSoftPosition', 'degree', rep=n)
        self.detector.angle_x = data_as(das, 'detectorArmAngle/softPosition', 'degree', rep=n)
        self.detector.angle_x_target = data_as(das, 'detectorArmAngle/desiredSoftPosition', 'degree', rep=n)
        # CRUFT: renamed detectorTableMotor to detectorArmAngle
        if self.detector.angle_x is None:
            self.detector.angle_x = data_as(das, 'detectorTableMotor/softPosition', 'degree', rep=n)
        if self.detector.angle_x_target is None:
            self.detector.angle_x_target = data_as(das, 'detectorTableMotor/desiredSoftPosition', 'degree', rep=n)
        # PAK 2020-11-11: detectorTable/rowAngularOffsets changed orientation; flatten so we don't care
        self.detector.angle_x_offset = data_as(das, 'detectorTable/rowAngularOffsets', 'degree').flatten()

        # Attenuators
        self.attenuator.transmission = data_as(entry, 'instrument/attenuator/transmission', '', dtype="float")
        self.attenuator.transmission_err = data_as(entry, 'instrument/attenuator/transmission_err', '', dtype="float")
        self.attenuator.target_value = data_as(das, 'counter/actualAttenuatorsDropped', '', rep=n)

        #print("shapes", self.detector.counts.shape, self.detector.wavelength.shape, self.detector.efficiency.shape)
        #print("shapes", self.sample.angle_x.shape, self.detector.angle_x.shape, self.detector.angle_x_offset.shape)


    @property
    def Ti(self):
        return self.sample.angle_x[:, None, None]

    @property
    def Ti_target(self):
        return self.sample.angle_x_target[:, None, None]

    @property
    def Td(self):
        angle, offset = self.detector.angle_x, self.detector.angle_x_offset
        return angle[:, None, None] + offset[None, None, :]

    @property
    def Td_target(self):
        angle, offset = self.detector.angle_x_target, self.detector.angle_x_offset
        return angle[:, None, None] + offset[None, None, :]

    @property
    def Li(self):
        # Assume incident = reflected wavelength if no monochromator
        return (self.Ld if self.monochromator.wavelength is None
                else self.monochromator.wavelength)

    @property
    def Ld(self):
        return self.detector.wavelength  # Candor always has detector wavelength
        #return (self.Li if self.detector.wavelength is None
        #        else self.detector.wavelength)

    def plot(self, label=None):
        if label is None:
            label = self.name+self.polarization

        from matplotlib import pyplot as plt
        #v = np.log10(self.v + (self.v == 0))
        # set noise floor to ignore 5% of the data above zero
        v = self.v.copy()
        vmin = np.sort(v[v > 0])[int((v > 0).size * 0.01)]
        v[v<vmin] = vmin
        v = np.log10(v)

        channel = np.arange(1, NUM_CHANNELS+1)

        if True: # Qz - channel OR S1 - channel
            if Intent.isslit(self.intent):
                x, xlabel = self.slit1.x, "Slit 1 opening (mm)"
                x = x[:, None, None]
                x = np.concatenate((x, x), axis=2)
            else:
                x, xlabel = self.Qz, "Qz (1/Ang)"
            y, ylabel = np.vstack((channel, channel)).T[None, :, :], "channel"
            #print("Qz-channel", x.shape, y.shape, v.shape)

            #plt.figure()
            plt.subplot(211)
            plt.pcolormesh(x[:, :, 0].T, y[:, :, 0].T, v[:, :, 0].T)
            plt.xlabel(xlabel)
            plt.ylabel("{ylabel} - bank 0".format(ylabel=ylabel))
            plt.colorbar()
            plt.subplot(212)
            plt.pcolormesh(x[:, :, 1].T, y[:, :, 1].T, v[:, :, 1].T)
            plt.xlabel(xlabel)
            plt.ylabel("{ylabel} - bank 1".format(ylabel=ylabel))
            plt.colorbar()
            plt.suptitle('detector counts for {self.name}'.format(self=self))

        if False: # lambda-theta
            x, xlabel = self.detector.wavelength, 'detector wavelength (Ang)'
            y, ylabel = self.sample.angle_x, 'sample angle (degree)'
            #print("lambda-theta", x.shape, y.shape, v.shape)

            plt.figure()
            plt.subplot(211)
            plt.pcolormesh(edges(x[0, :, 0]), edges(y), v[:, :, 0])
            plt.xlabel(xlabel)
            plt.ylabel("{ylabel} - bank 0".format(ylabel=ylabel))
            plt.colorbar()
            plt.subplot(212)
            plt.pcolormesh(edges(x[0, :, 1]), edges(y), v[:, :, 1])
            plt.xlabel(xlabel)
            plt.ylabel("{ylabel} - bank 1".format(ylabel=ylabel))
            plt.colorbar()
            plt.suptitle('detector counts for {self.name}'.format(self=self))

        if False: # detector efficiency
            eff = self.detector.efficiency[0]
            #print("detector efficiency", eff.shape)

            plt.figure()
            plt.plot(channel, eff[:, 0], label='bank 0')
            plt.plot(channel, eff[:, 1], label='bank 1')
            plt.xlabel('channel number')
            plt.ylabel('wavelength efficiency')
            plt.suptitle('detectorTable.detectorEfficiencies for {self.name}'.format(self=self))
            plt.legend()

        if False: # detector wavelength
            y, ylabel = self.detector.wavelength, 'detector wavelength (Ang)'
            dy = self.detector.wavelength_resolution
            #print("detector wavelength", y.shape)

            plt.figure()
            plt.errorbar(channel, y[0, :, 0], yerr=dy[0, :, 0], label='bank 0')
            plt.errorbar(channel, y[0, :, 1], yerr=dy[0, :, 1], label='bank 1')
            plt.xlabel('channel number')
            plt.ylabel(ylabel)
            plt.suptitle('detectorTable.wavelength for {self.name}'.format(self=self))
            plt.legend()

    def get_axes(self):
        x, xlabel = self.detector.wavelength, "Wavelength (Ang)"
        #print("intent", self.intent)
        #if Intent.isslit(self.intent):
        #    y, ylabel = self.slit1.x, "S1 (mm)"
        #elif Intent.isspec(self.intent):
        #    y, ylabel = self.sample.angle_x, "Detector Angle (degrees)"
        if self.detector.angle_x.max() - self.detector.angle_x.min() > 0.01:
            y, ylabel = self.detector.angle_x, "Detector Angle (degrees)"
        elif self.sample.angle_x.max() - self.sample.angle_x.min() > 0.01:
            y, ylabel = self.sample.angle_x, "Sample Angle (degrees)"
        elif self.slit1.x.max() - self.slit1.x.min() > 0.01:
            y, ylabel = self.slit1.x, "S1 (mm)"
        else:
            y, ylabel = np.arange(1, len(self.v)+1), "point"
        return (x, xlabel), (y, ylabel)

    def get_plottable(self):
        name = getattr(self, "name", "default_name")
        entry = getattr(self, "entry", "default_entry")
        def limits(vec, n):
            low, high = vec.min(), vec.max()
            delta = (high - low) / max(n-1, 1)
            # TODO: move range cleanup to plotter
            if delta == 0.:
                delta = vec[0]/10.
            return low - delta/2, high+delta/2
        data = self.v
        ny, nx, nbanks = data.shape  # nangles, nwavelengths, nbanks
        (x, xlabel), (y, ylabel) = self.get_axes()
        x, nx = np.array([x.min(), x.min() + nbanks*(x.max() - x.min())]), nbanks*nx
        xmax, xmin = limits(x, nx) # reversed because wavelengths are reversed
        ymin, ymax = limits(y, ny)
        # TODO: self.detector.mask
        zmin, zmax = data.min(), data.max()
        # TODO: move range cleanup to plotter
        if zmin <= 0.:
            if (data > 0).any():
                zmin = data[data > 0].min()
            else:
                data[:] = zmin = 1e-10
            if zmin >= zmax:
                zmax = 10*zmin
        dims = {
            "xmin": xmin, "xmax": xmax, "xdim": nx,
            "ymin": ymin, "ymax": ymax, "ydim": ny,
            "zmin": zmin, "zmax": zmax,
        }
        # One z per detector bank
        #z = [data[..., k].ravel('F').tolist() for k in range(data.shape[-1])]
        # One z with banks back-to-back
        z = [data.ravel('F').tolist()]
        #print("data", data.shape, dims, len(z))
        plottable = {
            'type': '2d_multi',
            'dims': {'zmin': zmin, 'zmax': zmax},
            'datasets': [{'dims': dims, 'data': z[0]}],
            'entry': entry,
            'title': "%s:%s" % (name, entry),
            'options': {},
            'xlabel': xlabel,
            'ylabel': ylabel,
            'zlabel': 'Intensity (I)',
            'ztransform': 'log',
        }
        #print(plottable)
        return plottable

    # TODO: Define export format for partly reduced PSD data.
    @exports_json("json")
    def to_json_text(self):
        name = getattr(self, "name", "default_name")
        entry = getattr(self, "entry", "default_entry")
        return {
            "name": name,
            "entry": entry,
            "file_suffix": ".json",
            "value": self._toDict(),
            }

    # Kill column writer
    def to_column_text(self):
        pass

class QData(ReflData):
    def __init__(self, data, q, dq, v, dv, ti=None, dt=None, ld=None, dl=None):
        super().__init__()
        self.probe = data.probe
        self.path = data.path
        self.uri = data.uri
        self.name = data.name
        self.entry = data.entry
        self.description = data.description
        self.polarization = data.polarization
        self.normbase = data.normbase
        self.v = v
        self.dv = dv
        self.intent = Intent.spec
        self.points = len(q)
        self.scan_value = []
        self.scan_units = []
        self.scan_label = []
        # Put angles and resolution in the appropriate places in the refldata
        # data structure.
        self.sample = copy(data.sample)
        self.sample.angle_x = ti
        self.sample.angle_x_target = ti
        self.angular_resolution = dt
        self.detector = Detector()
        self.detector.wavelength = ld
        self.detector.wavelength_resolution = dl
        self.detector.angle_x = 2*ti
        self.detector.angle_x_target = 2*ti
        # Since dq is not computed directly from dt and dl, we need to store
        # it specially. Need to override apply_mask to fix up dq.
        self._dq = dq

    @property
    def dQ(self):
        return self._dq

    def apply_mask(self, mask_indices):
        self._dq = np.delete(self._dq, mask_indices)
        ReflData.apply_mask(self, mask_indices)

def nobin(data):
    """
    Dump the raw data points into a single Q vector.
    """
    # Make all data have the same shape
    columns = (
        data.Qz, data.dQ, data.v, data.dv,
        data.Ti, data.angular_resolution, data.Ld, data.dL)
    columns = [np.broadcast_to(p, data.v.shape) for p in columns]

    # Process the detector banks into different data sets
    datasets = []
    for bank in range(data.v.shape[2]):
        columns_bank = [p[..., bank].flatten() for p in columns]
        index = np.argsort(columns_bank[0])
        columns_bank = [p[index] for p in columns_bank]
        Qx, dQ, v, dv, Ti, dT, L, dL = columns_bank
        datasets.append(QData(data, Qx, dQ, v, dv, Ti, dT, L, dL))

    # Only look at bank 0 for now
    return datasets[0]

def rebin(data, q, average="poisson"):
    if average not in ("poisson", "gauss"):
        raise ValueError("expected average to be 'poisson' or 'gauss'")
    if average == "poisson" and data.normbase not in ("monitor", "time", "none"):
        raise ValueError("expected norm to be time, monitor or none for poisson average")
    q_edges = edges(q, extended=True)
    datasets = []
    for bank in range(data.v.shape[2]):
        qz, dq, v, dv, Ti, dT, L, dL = _rebin_bank(data, bank, q_edges, average)
        #output = ReflData()
        #output.v, output.dv = v, dv
        #output.x, output.dx = q, dq
        datasets.append(QData(data, qz, dq, v, dv, Ti, dT, L, dL))

    # Only look at bank 0 for now
    return datasets[0]

def _rebin_bank(data, bank, q_edges, average):
    """
    Merge q points across channels and angles, returning q, dq, v, dv.

    Intensities (v, dv) are combined using poisson averaging.

    Q values (Q, dQ) for the combined measurements are weighted by intensity.
    This means that identical measurement conditions may give different
    (Q, dQ) depending on the specific values measured.

    The following are computed from the combined points::

        [q] = <q> = <4π/λ sin(θ)>
        [λ] = 1/<1/λ>
        [θ] = arcsin([q] [λ] / 4π)
        [Δq]² = <q √ {(Δλ/λ)² + (Δθ/tan θ)²}> + <q²> - <q>²
        [Δλ]² = <Δλ²> + <λ²> - <λ>²
        [Δθ]² = <Δθ²>
        [q'] = 4π/[λ] sin([θ] + δ)                     for θ-offset δ
        [Δq']² = [Δq]² + ([q]/tan[θ])² (2ω [Δθ] + ω²)  for sample broadening ω

    The result for [q'] and [Δq'] are with 1% over a wide range of angles
    and slits. Only small θ-offset values were checked since large errors
    in incident angle are readily visible in the data. The reflectivity curves
    will be clearly misaligned especially near the critical edge for θ-offset
    above about 0.1°. Large values of sample broadening are supported, with
    up to 2° tested. Negative sample broadening will lead to anomalies at low
    angles.

    The <q²> - <q>² term in [Δq] comes from the formula for variance in a
    mixture distribution, which averages the variances of the individual
    distributions and adds a spread term in case the means are not overlapping.
    See <https://en.wikipedia.org/wiki/Mixture_distribution#Moments>_.

    The sample broadening formula [Δq'] comes from substituting Δθ+ω for Δθ
    in [Δq] and expanding the square. By using [Δq]² to compute [Δq']², the
    spread term is automatically incorporated. This change may require updates
    to the fitting software, which compute [Δq'] from (θ,λ,Δθ,Δλ) directly.

    Since θ and λ are completely correlated based on the q value each bin
    has thin resolution function following the constant q curve on a θ-λ plot.
    The resulting Δq is smaller than would be expected from the full Δθ and Δλ
    within the bin.

    Since the distribution is not a simple gaussian, we are free to choose
    [θ], [λ], [Δθ] and [Δλ] in a way that is convenient for fitting.  Choosing
    wavelength based on the average of the inverse, and angle so that it
    matches the corresponding [q] works well for computing θ-offset.
    Using this [θ] and averaging the variance for [Δθ] works well for
    estimating the effects of sample broadening.

    The [Δλ] term is set from the second central moment from the mixture
    distribution. This gives some idea of the range of wavelengths included
    in each [q], but it is not not directly useful. To properly fit data in
    which reflectivity is wavelength dependent the correct wavelength
    distribution is needed, not just the first and second moment. Because
    points with different intensity are combined, even knowing that incident
    wavelength follows a truncated Maxwell-Boltzmann distribution with a
    given temperature is not enough to reconstruct the measured distribution.
    In these situations measure fewer angles for longer without binning the
    data so that you can ignore wavelength variation within each theory value.
    """
    # Make all data have the same shape
    columns = (
        data.Qz, data.dQ, data.v, data.dv,
        data.Ti, data.angular_resolution, data.Ld, data.dL)
    columns = [np.broadcast_to(p, data.v.shape) for p in columns]
    columns = [p[:, :, bank].flatten() for p in columns]
    q, dq, y, dy, T, dT, L, dL = columns

    # Sort q values into bins
    nbins = len(q_edges) - 1
    bin_index = np.searchsorted(q_edges, q) - 1

    # Some bins may not have any points contributing, such as those before
    # and after, or those in the middle if the q-step is too fine. These
    # will be excluded from the final result.
    points_per_bin = np.bincount(bin_index, minlength=nbins)
    # Note: we add empty_q to the divisor in a number of places to protect
    # against divide by zero in those bins. Since we are excluding these at
    # the end, this removes the spurious warnings without changing results.
    empty_q = (points_per_bin == 0)
    # The following is cribbed from util.poisson_average, replacing
    # np.sum with np.bincount.
    # TODO: update poisson average so it handles grouping
    norm = data.normbase
    if average == "gauss":
        dy = dy + (dy == 0) # protect against zero uncertainty
        Swx = np.bincount(bin_index, weights=y/dy**2, minlength=nbins)
        Sw = np.bincount(bin_index, weights=dy**-2, minlength=nbins)
        Sw += empty_q  # Protect against division by zero
        bar_y = Swx / Sw
        bar_dy = 1/np.sqrt(Sw)
    elif norm == "none":
        bar_y = np.bincount(bin_index, weights=y, minlength=nbins)
        bar_dy = np.sqrt(np.bincount(bin_index, weights=dy**2, minlength=nbins))
    else:
        # Counts must be positive for poisson averaging...
        y = y.copy()
        y[y < 0] = 0.
        dy = dy + (dy == 0) # protect against zero uncertainty
        monitors = y*(y+1)/dy**2 if norm == "monitor" else y/dy**2 # if "time"
        monitors[y == 0] = 1./dy[y == 0] # protect against zero counts
        counts = y*monitors
        combined_monitors = np.bincount(bin_index, weights=monitors, minlength=nbins)
        combined_counts = np.bincount(bin_index, weights=counts, minlength=nbins)
        combined_monitors += empty_q  # Protect against division by zero
        bar_y = combined_counts/combined_monitors
        if norm == "time":
            bar_dy = np.sqrt(bar_y / combined_monitors)
        else:
            bar_dy = 1./combined_monitors * np.sqrt(1. + 1./combined_monitors)
            idx = (bar_y != 0)
            bar_dy[idx] = bar_y[idx] * np.sqrt(1./combined_counts[idx]
                                               + 1./combined_monitors[idx])

    # Find Q center and resolution, weighting by intensity
    w = np.ones_like(y)  # Weights must be positive; use equal weights for now
    #w = y # use intensity weighting when finding q centers
    sum_w = np.bincount(bin_index, weights=w, minlength=nbins)
    #assert ((sum_w == 0) == empty_q).all()
    sum_w += empty_q  # protect against divide by zero
    sum_q = np.bincount(bin_index, weights=q*w, minlength=nbins)
    bar_q = sum_q / sum_w
    # Combined dq according to mixture distribution.
    sum_dqsq = np.bincount(bin_index, weights=w*(dq**2 + q**2), minlength=nbins)
    bar_dq = np.sqrt(sum_dqsq/sum_w - bar_q**2)
    ## Combined dq according average of variance.
    #bar_dq = np.sqrt(np.bincount(bin_index, weights=dq*w, minlength=nbins)/bar_w)
    # Set dq to 1% for now...
    #bar_dq = bar_q*0.01

    # Combine wavelengths
    sum_Linv = np.bincount(bin_index, weights=w/L, minlength=nbins)
    #assert ((sum_Linv == 0) == empty_q).all()
    sum_Linv += empty_q  # protect against divide by zero
    bar_Linv = sum_w/sum_Linv  # Not the first moment of L
    sum_L = np.bincount(bin_index, weights=w*L, minlength=nbins)
    sum_dLsq = np.bincount(bin_index, weights=w*(dL**2+L**2), minlength=nbins)
    bar_dL = np.sqrt(sum_dLsq/sum_w - (sum_L/sum_w)**2)

    # Combine angles
    bar_T = np.degrees(np.arcsin(bar_q*bar_Linv / 4 / np.pi))
    sum_dT = np.bincount(bin_index, weights=w*dT**2, minlength=nbins)
    bar_dT = np.sqrt(sum_dT/sum_w)

    # Need to drop catch-all bins before and after q edges.
    # Also need to drop q bins which don't contain any values.
    keep = ~empty_q
    keep[0] = keep[-1] = False
    columns = (bar_q, bar_dq, bar_y, bar_dy, bar_T, bar_dT, bar_Linv, bar_dL)
    return [p[keep] for p in columns]

def edges(c, extended=False):
    r"""
    Linear bin edges given centers.

    If *extended* then create before/after bins so coverage is $(-\infty, \infty)$.
    """
    midpoints = (c[:-1]+c[1:])/2
    left = 2*c[0] - midpoints[0]
    right = 2*c[-1] - midpoints[-1]
    if extended:
        return np.hstack((-np.inf, left, midpoints, right, np.inf))
    else:
        return np.hstack((left, midpoints, right))

if __name__ == "__main__":
    from .nexusref import demo
    demo(loader=load_entries)



r"""
Resolution calculations
"""

import numpy as np
from numpy import pi, sqrt, log, degrees, radians, cos, sin, tan
from numpy import arcsin as asin, ceil
from numpy import ones_like, arange, isscalar, asarray, hstack

from . import util


def QL2T(Q=None, L=None):
    r"""
    Compute angle from $Q$ (|1/Ang|) and wavelength (|Ang|).

    .. math::

        \theta = \sin^{-1}( |Q| \lambda / 4 \pi )

    Returns $\theta$\ |deg|.
    """
    Q, L = asarray(Q, 'd'), asarray(L, 'd')
    return degrees(asin(abs(Q) * L / (4*pi)))


def TL2Q(T=None, L=None):
    r"""
    Compute $Q$ from angle (|deg|) and wavelength $L$ (|Ang|).

    .. math::

        Q = 4 \pi \sin(\theta) / \lambda

    Returns $Q$ |1/Ang|
    """
    T, L = radians(asarray(T, 'd')), asarray(L, 'd')
    return 4 * pi * sin(T) / L


_FWHM_scale = 2*sqrt(2*log(2))
def FWHM2sigma(s):
    return s/_FWHM_scale


def sigma2FWHM(s):
    return s*_FWHM_scale


def calc_Qx(Ti, Td, Li, Ld):
    # type: (np.ndarray, np.ndarray, np.ndarray) -> (np.ndarray, np.ndarray)
    ki, kf = 2*pi/Li, 2*pi/Ld
    Qx = ki*cos(radians(Td - Ti)) - kf*cos(radians(-Ti))
    return Qx


def calc_Qz(Ti, Td, Li, Ld):
    ki, kf = 2*pi/Li, 2*pi/Ld
    Qz = ki*sin(radians(Td - Ti)) - kf*sin(radians(-Ti))
    return Qz


def TiTdL2Qxz(Ti, Td, L):
    # type: (np.ndarray, np.ndarray, np.ndarray) -> (np.ndarray, np.ndarray)
    Qx = 2*pi/L * (cos(radians(Td - Ti)) - cos(radians(Ti)))
    Qz = 2*pi/L * (sin(radians(Td - Ti)) + sin(radians(Ti)))
    return Qx, Qz


def dTdL2dQ(T=None, dT=None, L=None, dL=None):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray) -> np.ndarray
    r"""
    Convert wavelength dispersion and angular divergence to $Q$ resolution.

    *T*, *dT*  (|deg|) angle and 1-\ $\sigma$ angular divergence
    *L*, *dL*  (|Ang|) wavelength and 1-\ $\sigma$ wavelength dispersion

    Returns 1-\ $\sigma$ $\Delta Q$

    Given $Q = 4 \pi sin(\theta)/\lambda$, this follows directly from
    gaussian error propagation using

    ..math::

        \Delta Q^2
            &= \left(\frac{\partial Q}{\partial \lambda}\right)^2\Delta\lambda^2
            + \left(\frac{\partial Q}{\partial \theta}\right)^2\Delta\theta^2

            &= Q^2 \left(\frac{\Delta \lambda}{\lambda}\right)^2
            + Q^2 \left(\frac{\Delta \theta}{\tan \theta}\right)^2

            &= Q^2 \left(\frac{\Delta \lambda}{\lambda}\right)^2
            + \left(\frac{4\pi\cos\theta\,\Delta\theta}{\lambda}\right)^2

    with the final form chosen to avoid cancellation at $Q=0$.
    """

    # Compute dQ from wavelength dispersion (dL) and angular divergence (dT)
    T, dT = radians(asarray(T, 'd')), radians(asarray(dT, 'd'))
    L, dL = asarray(L, 'd'), asarray(dL, 'd')
    #print T, dT, L, dL
    dQ = (4*pi/L) * sqrt((sin(T)*dL/L)**2 + (cos(T)*dT)**2)

    #sqrt((dL/L)**2+(radians(dT)/tan(radians(T)))**2)*probe.Q
    return dQ


def dQ_broadening(dQ, L, T, dT, width):
    r"""
    Broaden an existing dQ by the given divergence.

    *dQ* |1/Ang|, with 1-\ $\sigma$ $Q$ resolution
    *L* |Ang|
    *T*, *dT* |deg|, with 1-\ $\sigma$ angular divergence
    *width* |deg|, with 1-\ $\sigma$ increased angular divergence

    The calculation is derived by substituting
    $\Delta\theta' = \Delta\theta + \omega$ for sample broadening $\omega$.
    """
    T, dT = radians(asarray(T, 'd')), radians(asarray(dT, 'd'))
    width = radians(width)
    dQsq = dQ**2 + (4*pi/L*cos(T))**2*(2*width*dT + width**2)

    return sqrt(dQsq)


def dQdT2dLoL(Q, dQ, T, dT):
    r"""
    Convert a calculated Q resolution and angular divergence to a
    wavelength dispersion.

    *Q*, *dQ* |1/Ang|, with 1-\ $\sigma$ $Q$ resolution
    *T*, *dT* |deg|, with 1-\ $\sigma$ angular divergence

    Returns 1-\ $\sigma$ $\Delta\lambda/\lambda$
    """
    T, dT = radians(asarray(T, 'd')), radians(asarray(dT, 'd'))
    Q, dQ = asarray(Q, 'd'), asarray(dQ, 'd')
    return sqrt((dQ/Q)**2 - (dT/tan(T))**2)


def dQdL2dT(Q, dQ, L, dL):
    r"""
    Convert a calculated Q resolution and wavelength dispersion to
    angular divergence.

    *Q*, *dQ* |1/Ang|, with 1-\ $\sigma$ $Q$ resolution
    *L*, *dL* |deg|, with 1-\ $\sigma$ angular divergence

    Returns FWHM $\theta, \Delta\theta$
    """
    L, dL = asarray(L, 'd'), asarray(dL, 'd')
    Q, dQ = asarray(Q, 'd'), asarray(dQ, 'd')
    T = QL2T(Q, L)
    dT = degrees(sqrt((dQ/Q)**2 - (dL/L)**2) * tan(radians(T)))
    return dT


Plancks_constant = 6.62618e-27 # Planck constant (erg*sec)
neutron_mass = 1.67495e-24     # neutron mass (g)
def TOF2L(d_moderator, TOF):
    r"""
    Convert neutron time-of-flight to wavelength.

    .. math::

        \lambda = (t/d) (h/n_m)

    where:

        | $\lambda$ is wavelength in |Ang|
        | $t$ is time-of-flight in $u$\s
        | $h$ is Planck's constant in erg seconds
        | $n_m$ is the neutron mass in g
    """
    return TOF*(Plancks_constant/neutron_mass/d_moderator)


def bins(low, high, dLoL):
    r"""
    Return bin centers from low to high preserving a fixed resolution.

    *low*, *high* are the minimum and maximum wavelength.
    *dLoL* is the desired resolution FWHM $\Delta\lambda/\lambda$ for the bins.
    """

    step = 1 + dLoL
    n = ceil(log(high/low)/log(step))
    edges = low*step**arange(n+1)
    L = (edges[:-1]+edges[1:])/2
    return L

def binwidths(L):
    r"""
    Determine the wavelength dispersion from bin centers *L* for time-of-flight
    measurements.

    The wavelength dispersion $\Delta\lambda$ is just the difference
    between consecutive bin edges, so:

    .. math::

        \Delta L_i  = E_{i+1}-E_{i}
                    = (1+\omega) E_i - E_i
                    = \omega E_i
                    = \frac{2 \omega}{2+\omega} L_i

    where $E$ and $\omega$ are as defined in :func:`binedges`.
    """
    if L[1] > L[0]:
        dLoL = L[1]/L[0] - 1
    else:
        dLoL = L[0]/L[1] - 1
    dL = 2*dLoL/(2+dLoL)*L
    return dL

def binedges(L):
    # type: (np.ndarray) -> np.ndarray
    r"""
    Construct bin edges *E* from bin centers *L* for time-of-flight
    measurements.

    Assuming fixed $\omega = \Delta\lambda/\lambda$ in the bins, the
    edges will be spaced logarithmically at:

    .. math::

        E_0     &= \min \lambda \\
        E_{i+1} &= E_i + \omega E_i = E_i (1+\omega)

    with centers $L$ half way between the edges:

    .. math::

        L_i = (E_i+E_{i+1})/2
            = (E_i + E_i (1+\omega))/2
            = E_i (2 + \omega)/2

    Solving for $E_i$, we can recover the edges from the centers:

    .. math::

        E_i = L_i \frac{2}{2+\omega}

    The final edge, $E_{n+1}$, does not have a corresponding center
    $L_{n+1}$ so we must determine it from the previous edge $E_n$:

    .. math::

        E_{n+1} = L_n \frac{2}{2+\omega}(1+\omega)

    The fixed $\omega$ can be retrieved from the ratio of any pair
    of bin centers using:

    .. math::

        \frac{L_{i+1}}{L_i} = \frac{ (E_{i+2}+E_{i+1})/2 }{ (E_{i+1}+E_i)/2 }
                          = \frac{ (E_{i+1}(1+\omega)+E_{i+1} }
                                  { (E_i(1+\omega)+E_i }
                          = \frac{E_{i+1}}{E_i}
                          = \frac{E_i(1+\omega)}{E_i} = 1 + \omega
    """
    if L[1] > L[0]:
        dLoL = L[1]/L[0] - 1.
        last = (1.+dLoL)
    else:
        dLoL = L[0]/L[1] - 1.
        last = 1./(1.+dLoL)
    E = L*(2./(2.+dLoL))
    return hstack((E, E[-1]*last))

def divergence(slits=None, distance=None, T=None,
               sample_width=np.inf, use_sample=True):
    # type: (Sequence[Union[float, np.ndarray], ...], Sequence[float, ...], Optional[np.ndarray], float, bool) -> np.ndarray
    r"""
    Calculate divergence due to slit and sample geometry.

    :Parameters:
        *slits*     : float OR (float, ...) | mm
            s1,s2,s3,s4 slit openings for all available slits
        *distance*  : (float, ...) | mm
            d1,d2,d3,d4 distance to each slit
        *T*         : float OR [float] | |deg|
            incident angles
        *sample_width*      : float | mm
            w, width of the sample
        *use_sample* : bool
            True if sample footprint should be treated as a slit.  If False,
            then incident angle T and sample width are not used.

    :Returns:
        *dT*  : float OR [float] | |deg| 1-\ $\sigma$
            calculated angular divergence

    **Algorithm:**

    The divergence is based on the slit openings and the distance between
    the slits.  For very small samples, where the slit opening is larger
    than the width of the sample across the beam, the sample itself acts
    like the second slit.

    First find $p$, the projection of the beam on the sample:

    .. math::

        p &= w \sin\left(\frac{\pi}{180}\theta\right)

    Define this as slit s0 at distance d0=0.

    Given a set of slit openings $s_i$ with distance from sample $d_i$,
    the $\Delta\theta_d$ full width at half maximum (FWHM) in radians is
    computed as:

    .. math::

        \Delta\theta_d = \min_{i \ne j) \frac{1}{2}\frac{s_i+s_j}{|d_i-d_j|}

    In addition to the slit divergence, we need to add in any sample
    broadening $\Delta\theta_s$ returning the total divergence in degrees:

    .. math::

        \Delta\theta = \frac{180}{\pi} \Delta\theta_d + \Delta\theta_s

    Reversing this equation, the sample broadening contribution can
    be measured from the FWHM of the rocking curve, $B$, measured in
    degrees at a particular angle and slit opening:

    .. math::

        \Delta\theta_s = B - \frac{180}{\pi}\Delta\theta_d

    If any of the slits is not centered or the sample is offset from the
    center of rotation then this calculation will not be correct.

    Depending on the slit openings and distances, the angular distribution
    in the beam from a set of slits will vary from triangular
    ($\sigma^2 = w^2/24$) to uniform ($\sigma^2 = w^2/12) where $w$ is the
    width of the base, as determined by the minimum and maximum achievable
    angles through all the slits. Any slits that are not centered on the
    beam path can introduce a skew in this distribution.

    The algorithm computes the symmetric trapezoidal distribution
    for each pair of slits along the beam path and uses the minimum divergence
    as the divergence of the complete beam.  Monte Carlo simulations show that
    this method underestimates the divergence by up to 30% when slit 1 is
    much more open than slit 2, but handles the other conditions very well.
    Note that the sample itself can act like slit 2 at low angles, as
    can slit 3.

    **TODO**: fix S1 >> S2 condition

    **TODO**: sample footprint not used even when *use_sample* is True
    """
    # Extend all arrays to look like T (for use in candor data)
    slits = [util.extend(s, T) for s in slits]
    distance = [util.extend(d, T) for d in distance]

    # Add sample projection at distance zero
    if False and use_sample and np.isfinite(sample_width):
        sample_projection = sample_width * abs(sin(radians(T)))
        slits = slits + [sample_projection]
        distance = distance + [0.]

    # Compute pair-wise delta-theta
    def _divergence(i, j):
        # Note: divergence_simple below uses (s1 + s2)/|2d| for FWHM. Given
        # s1 = s2, this form returns arctan((s1 + s2)/|2d|)/sqrt(6) 1-sigma.
        # sqrt(6) ~ 2.45 whereas FWHM to 1-sigma conversion ~ 2.35. The arctan
        # makes the values closer still.
        s1, s2, d = slits[i], slits[j], distance[i] - distance[j]
        w, t = np.arctan(abs((s1 + s2)/2/d)), np.arctan(abs((s1 - s2)/2/d))
        return np.sqrt((w**2 + t**2)/6) # 1-sigma radians
    n = len(slits)
    sigma_rad = [_divergence(i, j) for i in range(n) for j in range(i+1, n)]
    ## With four slit pairs plus sample, show whether the sample footprint
    ## is the limiting factor in the divergence. The structure of the
    ## divergence list sigma is as follows:
    ##  [S1:S2 S1:S3 S1:S4 S1:foot S2:S3 S2:S4 S2:foot S3:S4 S3:foot S4:foot]
    #print(f"sample dT = {min(sigma_rad[k].min() for k in (3, 6, 8, 9))},"
    #      f" slit dT = {min(sigma_rad[k].min() for k in (0, 1, 2, 4, 5, 7))}")

    # Find the minimum delta-theta across all pairs
    # Sample slit is one per angle, so some divergence values will be vectors
    sigma_rad = np.asarray(np.broadcast_arrays(*sigma_rad))
    min_sigma_rad = np.min(sigma_rad, axis=0)

    # Convert to 1-sigma distribution in degrees
    return degrees(min_sigma_rad)

def divergence_simple(slits=None, distance=None, T=None,
                      sample_width=np.inf, use_sample=True):
    # TODO: check that the formula is correct for T=0 => dT = s1 / d1
    # TODO: add sample_offset and compute full footprint
    d1, d2 = distance
    try:
        s1, s2 = slits
    except TypeError:
        s1 = s2 = slits

    # Compute FWHM angular divergence dT from the slits in degrees
    dT_s1_s2 = 0.5*(s1+s2)/abs(d1-d2)
    # Sample can act as a slit according to the projection of its width
    if use_sample and np.isfinite(sample_width):
        sample = sample_width * abs(sin(radians(T)))
        dT_s1_sample = 0.5*(s1+sample)/abs(d1)
        dT_s2_sample = 0.5*(s2+sample)/abs(d2)
        dT = np.minimum(dT_s1_s2, np.minimum(dT_s1_sample, dT_s2_sample))
        return FWHM2sigma(degrees(dT))
    else:
        return FWHM2sigma(degrees(dT_s1_s2))


def slit_widths(T=None, slits_at_Tlo=None, Tlo=90, Thi=90,
                slits_below=None, slits_above=None):
    """
    Compute the slit widths for the standard scanning reflectometer
    fixed-opening-fixed geometry.

    :Parameters:
        *T* : [float] | |deg|
            Specular measurement angles.
        *Tlo*, *Thi* : float | |deg|
            Start and end of the opening region.  The default if *Tlo* is
            not specified is to use fixed slits at *slits_below* for all
            angles.
        *slits_below*, *slits_above* : float OR [float,float] | mm
            Slits outside opening region.  The default is to use the
            values of the slits at the ends of the opening region.
        *slits_at_Tlo* : float OR [float,float] | mm
            Slits at the start of the opening region.

    :Returns:
        *s1*, *s2* : [float] | mm
            Slit widths for each theta.

    Slits are assumed to be fixed below angle *Tlo* and above angle *Thi*,
    and opening at a constant dT/T between them.

    Slit openings are defined by a tuple (s1,s2) or constant s=s1=s2.
    With no *Tlo*, the slits are fixed with widths defined by *slits_below*,
    which defaults to *slits_at_Tlo*.  With no *Thi*, slits are continuously
    opening above *Tlo*.

    .. Note::
         This function works equally well if angles are measured in
         radians and/or slits are measured in inches.

    """

    # Slits at T<Tlo
    if slits_below is None:
        slits_below = slits_at_Tlo
    try:
        b1, b2 = slits_below
    except TypeError:
        b1 = b2 = slits_below
    s1 = ones_like(T) * b1
    s2 = ones_like(T) * b2

    # Slits at Tlo<=T<=Thi
    try:
        m1, m2 = slits_at_Tlo
    except TypeError:
        m1 = m2 = slits_at_Tlo
    idx = abs(T) >= Tlo
    s1[idx] = m1 * T[idx]/Tlo
    s2[idx] = m2 * T[idx]/Tlo

    # Slits at T > Thi
    if slits_above is None:
        slits_above = m1 * Thi/Tlo, m2 * Thi/Tlo
    try:
        t1, t2 = slits_above
    except TypeError:
        t1 = t2 = slits_above
    idx = abs(T) > Thi
    s1[idx] = t1
    s2[idx] = t2

    return s1, s2


'''
def resolution(Q=None,s=None,d=None,L=None,dLoL=None,Tlo=None,Thi=None,
               s_below=None, s_above=None,
               broadening=0, sample_width=1e10, sample_distance=0):
    """
    Compute the resolution for Q on scanning reflectometers.

    broadening is the sample warp contribution to angular divergence, as
    measured by a rocking curve.  The value should be w - (s1+s2)/(2*d)
    where w is the full-width at half maximum of the rocking curve.

    For itty-bitty samples, provide a sample width w and sample distance ds
    from slit 2 to the sample.  If s_sample = sin(T)*w is smaller than s2
    for some T, then that will be used for the calculation of dT instead.

    """
    T = QL2T(Q=Q,L=L)
    slits = slit_widths(T=T, s=s, Tlo=Tlo, Thi=Thi)
    dT = divergence(T=T, slits=slits, sample_width=sample_width,
                    sample_distance=sample_distance) + broadening
    Q,dQ = Qresolution(L, dLoL*L, T, dT)
    return FWHM2sigma(dQ)

def demo():
    import pylab
    from numpy import linspace, exp, real, conj, sin, radians
    # Values from volfrac example in garefl
    T = linspace(0,9,140)
    Q = 4*pi*sin(radians(T))/5.0042
    dQ = resolution(Q,s=0.21,Tlo=0.35,d=1890.,L=5.0042,dLoL=0.009)
    #pylab.plot(Q,dQ)

    # Fresnel reflectivity for silicon
    rho,sigma=2.07,5
    kz=Q/2
    f = sqrt(kz**2 - 4*pi*rho*1e-6 + 0j)
    r = (kz-f)/(kz+f)*exp(-2*sigma**2*kz*f)
    r[abs(kz)<1e-10] = -1
    R = real(r*conj(r))
    pylab.errorbar(Q,R,xerr=dQ,fmt=',r',capsize=0)
    pylab.grid(True)
    pylab.semilogy(Q,R,',b')

    pylab.show()

def demo2():
    import pylab
    Q,R,dR = pylab.loadtxt('ga128.refl.mce').T
    dQ = resolution(Q, s=0.154, Tlo=0.36, d=1500., L=4.75, dLoL=0.02)
    pylab.errorbar(Q,R,xerr=dQ,yerr=dR,fmt=',r',capsize=0)
    pylab.grid(True)
    pylab.semilogy(Q,R,',b')
    pylab.show()



if __name__ == "__main__":
    demo2()
'''



# This program is public domain
# Author: Paul Kienzle
"""
Format values and uncertainties nicely for printing.

The formatted value uses only the number of digits warranted by
the uncertainty in the measurement.

:func:`format_value` shows the value without the uncertainty.

:func:`format_uncertainty_pm` shows the expanded format v +/- err.

:func:`format_uncertainty_compact` shows the compact format v(##),
where the number in parenthesis is the uncertainty in the last two digits of v.

:func:`format_uncertainty` uses the compact format by default, but this
can be changed to use the expanded +/- format by setting
format_uncertainty.compact to False.  This is a global setting which should
be considered a user preference.  Any library code that depends on a specific
format style should use the corresponding formatting function.

If the uncertainty is 0 or not otherwise provided, the simple
%g floating point format option is used.

Infinite and indefinite numbers are represented as inf and NaN.

Example::

    >>> v,dv = 757.2356,0.01032
    >>> print(format_uncertainty_pm(v,dv))
    757.236 +/- 0.010
    >>> print(format_uncertainty_compact(v,dv))
    757.236(10)
    >>> format_uncertainty.compact = False
    >>> print(format_uncertainty(v,dv))
    757.236 +/- 0.010
    >>> format_uncertainty.compact = True
    >>> print(format_uncertainty(v,dv))
    757.236(10)
"""
from __future__ import division
import math

from numpy import isinf, isnan, inf, nan

try:
    #from typing import Optional
    pass
except ImportError:
    pass

__all__ = ['format_value', 'format_uncertainty',
           'format_uncertainty_compact', 'format_uncertainty_pm',
          ]


# Coordinating scales across a set of numbers is not supported.  For easy
# comparison a set of numbers should be shown in the same scale.  One could
# force this from the outside by adding scale parameter (either 10**n, n, or
# a string representing the desired SI prefix) and having a separate routine
# which computes the scale given a set of values.

# Coordinating scales with units offers its own problems.  Again, the user
# may want to force particular units.  This can be done by outside of the
# formatting routines by scaling the numbers to the appropriate units then
# forcing them to print with scale 10**0.  If this is a common operation,
# however, it may want to happen inside.

# The value e<n> is currently formatted into the number.  Alternatively this
# scale factor could be returned so that the user can choose the appropriate
# SI prefix when printing the units.  This gets tricky when talking about
# composite units such as 2.3e-3 m**2 -> 2300 mm**2, and with volumes
# such as 1 g/cm**3 -> 1 kg/L.


def format_value(value, uncertainty):
    """
    Given *value* v and *uncertainty* dv, return a string v which is
    the value formatted with the appropriate number of digits.
    """
    return _format_uncertainty(value, uncertainty, compact=None)


def format_uncertainty_pm(value, uncertainty):
    """
    Given *value* v and *uncertainty* dv, return a string v +/- dv.
    """
    return _format_uncertainty(value, uncertainty, compact=False)


def format_uncertainty_compact(value, uncertainty):
    """
    Given *value* v and *uncertainty* dv, return the compact
    representation v(##), where ## are the first two digits of
    the uncertainty.
    """
    return _format_uncertainty(value, uncertainty, compact=True)


def format_uncertainty(value, uncertainty):
    """
    Value and uncertainty formatter.

    Either the expanded v +/- dv form or the compact v(##) form will be
    used depending on whether *format_uncertainty.compact* is True or False.
    The default is True.
    """
    return _format_uncertainty(value, uncertainty, format_uncertainty.compact)
format_uncertainty.compact = True

def _format_uncertainty(value, uncertainty, compact):
    # type: (float, Optional[float], Optional[bool]) -> str
    """
    Implementation of both the compact and the +/- formats.
    """
    # Handle indefinite value
    if isinf(value):
        return "inf" if value > 0 else "-inf"
    if isnan(value):
        return "NaN"

    # Handle indefinite uncertainty
    if uncertainty is None or uncertainty <= 0 or isnan(uncertainty):
        return "%g" % value
    if isinf(uncertainty):
        if compact is None:
            return "%.2g" % value
        elif compact:
            return "%.2g(inf)" % value
        else:
            return "%.2g +/- inf" % value

    # Handle zero and negative values
    sign = "-" if value < 0 else ""
    value = abs(value)

    # Determine scale of value and error
    err_place = int(math.floor(math.log10(uncertainty)))
    if value == 0:
        val_place = err_place - 1
    else:
        val_place = int(math.floor(math.log10(value)))

    if err_place > val_place:
        # Degenerate case: error bigger than value
        # The mantissa is 0.#(##)e#, 0.0#(##)e# or 0.00#(##)e#
        val_place = err_place + 2
    elif err_place == val_place:
        # Degenerate case: error and value the same order of magnitude
        # The value is ##(##)e#, #.#(##)e# or 0.##(##)e#
        val_place = err_place + 1
    elif err_place <= 1 and val_place >= -3:
        # Normal case: nice numbers and errors
        # The value is ###.###(##)
        val_place = 0
    else:
        # Extreme cases: zeros before value or after error
        # The value is ###.###(##)e#, ##.####(##)e# or #.#####(##)e#
        pass

    # Force engineering notation, with exponent a multiple of 3
    val_place = int(math.floor(val_place / 3.)) * 3

    # Format the result
    digits_after_decimal = abs(val_place - err_place + 1)
    # Only use one digit of uncertainty if no precision included in result
    #if compact is None: digits_after_decimal -= 1
    val_str = "%.*f" % (digits_after_decimal, value / 10. ** val_place)
    exp_str = "e%d" % val_place if val_place != 0 else ""
    if compact is None:
        result = "".join((sign, val_str, exp_str))
    elif compact:
        err_str = "(%2d)" % int(uncertainty / 10. ** (err_place - 1) + 0.5)
        result = "".join((sign, val_str, err_str, exp_str))
    else:
        err_str = "%.*f" % (digits_after_decimal,
                            uncertainty / 10. ** val_place)
        result = "".join((sign, val_str, exp_str + " +/- ", err_str, exp_str))
    # print sign,value, uncertainty, "=>", result
    return result


def test_compact():
    value_str = format_uncertainty_compact

    # val_place > err_place
    assert value_str(1235670, 766000) == "1.24(77)e6"
    assert value_str(123567., 76600) == "124(77)e3"
    assert value_str(12356.7, 7660) == "12.4(77)e3"
    assert value_str(1235.67, 766) == "1.24(77)e3"
    assert value_str(123.567, 76.6) == "124(77)"
    assert value_str(12.3567, 7.66) == "12.4(77)"
    assert value_str(1.23567, .766) == "1.24(77)"
    assert value_str(.123567, .0766) == "0.124(77)"
    assert value_str(.0123567, .00766) == "0.0124(77)"
    assert value_str(.00123567, .000766) == "0.00124(77)"
    assert value_str(.000123567, .0000766) == "124(77)e-6"
    assert value_str(.0000123567, .00000766) == "12.4(77)e-6"
    assert value_str(.00000123567, .000000766) == "1.24(77)e-6"
    assert value_str(.000000123567, .0000000766) == "124(77)e-9"
    assert value_str(.00000123567, .0000000766) == "1.236(77)e-6"
    assert value_str(.0000123567, .0000000766) == "12.357(77)e-6"
    assert value_str(.000123567, .0000000766) == "123.567(77)e-6"
    assert value_str(.00123567, .000000766) == "0.00123567(77)"
    assert value_str(.0123567, .00000766) == "0.0123567(77)"
    assert value_str(.123567, .0000766) == "0.123567(77)"
    assert value_str(1.23567, .000766) == "1.23567(77)"
    assert value_str(12.3567, .00766) == "12.3567(77)"
    assert value_str(123.567, .0764) == "123.567(76)"
    assert value_str(1235.67, .764) == "1235.67(76)"
    assert value_str(12356.7, 7.64) == "12356.7(76)"
    assert value_str(123567, 76.4) == "123567(76)"
    assert value_str(1235670, 764) == "1.23567(76)e6"
    assert value_str(12356700, 764) == "12.35670(76)e6"
    assert value_str(123567000, 764) == "123.56700(76)e6"
    assert value_str(123567000, 7640) == "123.5670(76)e6"
    assert value_str(1235670000, 76400) == "1.235670(76)e9"

    # val_place == err_place
    assert value_str(123567, 764000) == "0.12(76)e6"
    assert value_str(12356.7, 76400) == "12(76)e3"
    assert value_str(1235.67, 7640) == "1.2(76)e3"
    assert value_str(123.567, 764) == "0.12(76)e3"
    assert value_str(12.3567, 76.4) == "12(76)"
    assert value_str(1.23567, 7.64) == "1.2(76)"
    assert value_str(.123567, .764) == "0.12(76)"
    assert value_str(.0123567, .0764) == "12(76)e-3"
    assert value_str(.00123567, .00764) == "1.2(76)e-3"
    assert value_str(.000123567, .000764) == "0.12(76)e-3"

    # val_place == err_place-1
    assert value_str(123567, 7640000) == "0.1(76)e6"
    assert value_str(12356.7, 764000) == "0.01(76)e6"
    assert value_str(1235.67, 76400) == "0.001(76)e6"
    assert value_str(123.567, 7640) == "0.1(76)e3"
    assert value_str(12.3567, 764) == "0.01(76)e3"
    assert value_str(1.23567, 76.4) == "0.001(76)e3"
    assert value_str(.123567, 7.64) == "0.1(76)"
    assert value_str(.0123567, .764) == "0.01(76)"
    assert value_str(.00123567, .0764) == "0.001(76)"
    assert value_str(.000123567, .00764) == "0.1(76)e-3"

    # val_place == err_place-2
    assert value_str(12356700, 7640000000) == "0.0(76)e9"
    assert value_str(1235670, 764000000) == "0.00(76)e9"
    assert value_str(123567, 76400000) == "0.000(76)e9"
    assert value_str(12356, 7640000) == "0.0(76)e6"
    assert value_str(1235, 764000) == "0.00(76)e6"
    assert value_str(123, 76400) == "0.000(76)e6"
    assert value_str(12, 7640) == "0.0(76)e3"
    assert value_str(1, 764) == "0.00(76)e3"
    assert value_str(0.1, 76.4) == "0.000(76)e3"
    assert value_str(0.01, 7.64) == "0.0(76)"
    assert value_str(0.001, 0.764) == "0.00(76)"
    assert value_str(0.0001, 0.0764) == "0.000(76)"
    assert value_str(0.00001, 0.00764) == "0.0(76)e-3"

    # val_place == err_place-3
    assert value_str(12356700, 76400000000) == "0.000(76)e12"
    assert value_str(1235670, 7640000000) == "0.0(76)e9"
    assert value_str(123567, 764000000) == "0.00(76)e9"
    assert value_str(12356, 76400000) == "0.000(76)e9"
    assert value_str(1235, 7640000) == "0.0(76)e6"
    assert value_str(123, 764000) == "0.00(76)e6"
    assert value_str(12, 76400) == "0.000(76)e6"
    assert value_str(1, 7640) == "0.0(76)e3"
    assert value_str(0.1, 764) == "0.00(76)e3"
    assert value_str(0.01, 76.4) == "0.000(76)e3"
    assert value_str(0.001, 7.64) == "0.0(76)"
    assert value_str(0.0001, 0.764) == "0.00(76)"
    assert value_str(0.00001, 0.0764) == "0.000(76)"
    assert value_str(0.000001, 0.00764) == "0.0(76)e-3"

    # Zero values
    assert value_str(0, 7640000) == "0.0(76)e6"
    assert value_str(0, 764000) == "0.00(76)e6"
    assert value_str(0,  76400) == "0.000(76)e6"
    assert value_str(0,   7640) == "0.0(76)e3"
    assert value_str(0,    764) == "0.00(76)e3"
    assert value_str(0,     76.4) == "0.000(76)e3"
    assert value_str(0,      7.64) == "0.0(76)"
    assert value_str(0,      0.764) == "0.00(76)"
    assert value_str(0,      0.0764) == "0.000(76)"
    assert value_str(0,      0.00764) == "0.0(76)e-3"
    assert value_str(0,      0.000764) == "0.00(76)e-3"
    assert value_str(0,      0.0000764) == "0.000(76)e-3"

    # negative values
    assert value_str(-1235670, 765000) == "-1.24(77)e6"
    assert value_str(-1.23567, .766) == "-1.24(77)"
    assert value_str(-.00000123567, .0000000766) == "-1.236(77)e-6"
    assert value_str(-12356.7, 7.64) == "-12356.7(76)"
    assert value_str(-123.567, 764) == "-0.12(76)e3"
    assert value_str(-1235.67, 76400) == "-0.001(76)e6"
    assert value_str(-.000123567, .00764) == "-0.1(76)e-3"
    assert value_str(-12356, 7640000) == "-0.0(76)e6"
    assert value_str(-12, 76400) == "-0.000(76)e6"
    assert value_str(-0.0001, 0.764) == "-0.00(76)"

    # non-finite values
    assert value_str(-inf, None) == "-inf"
    assert value_str(inf, None) == "inf"
    assert value_str(nan, None) == "NaN"

    # bad or missing uncertainty
    assert value_str(-1.23567, nan) == "-1.23567"
    assert value_str(-1.23567, -inf) == "-1.23567"
    assert value_str(-1.23567, -0.1) == "-1.23567"
    assert value_str(-1.23567, 0) == "-1.23567"
    assert value_str(-1.23567, None) == "-1.23567"
    assert value_str(-1.23567, inf) == "-1.2(inf)"


def test_pm():
    value_str = format_uncertainty_pm

    # val_place > err_place
    assert value_str(1235670, 766000) == "1.24e6 +/- 0.77e6"
    assert value_str(123567., 76600) == "124e3 +/- 77e3"
    assert value_str(12356.7,  7660) == "12.4e3 +/- 7.7e3"
    assert value_str(1235.67,   766) == "1.24e3 +/- 0.77e3"
    assert value_str(123.567,    76.6) == "124 +/- 77"
    assert value_str(12.3567,     7.66) == "12.4 +/- 7.7"
    assert value_str(1.23567,      .766) == "1.24 +/- 0.77"
    assert value_str(.123567,      .0766) == "0.124 +/- 0.077"
    assert value_str(.0123567,     .00766) == "0.0124 +/- 0.0077"
    assert value_str(.00123567,    .000766) == "0.00124 +/- 0.00077"
    assert value_str(.000123567,   .0000766) == "124e-6 +/- 77e-6"
    assert value_str(.0000123567,  .00000766) == "12.4e-6 +/- 7.7e-6"
    assert value_str(.00000123567, .000000766) == "1.24e-6 +/- 0.77e-6"
    assert value_str(.000000123567, .0000000766) == "124e-9 +/- 77e-9"
    assert value_str(.00000123567, .0000000766) == "1.236e-6 +/- 0.077e-6"
    assert value_str(.0000123567,  .0000000766) == "12.357e-6 +/- 0.077e-6"
    assert value_str(.000123567,   .0000000766) == "123.567e-6 +/- 0.077e-6"
    assert value_str(.00123567,    .000000766) == "0.00123567 +/- 0.00000077"
    assert value_str(.0123567,     .00000766) == "0.0123567 +/- 0.0000077"
    assert value_str(.123567,      .0000766) == "0.123567 +/- 0.000077"
    assert value_str(1.23567,      .000766) == "1.23567 +/- 0.00077"
    assert value_str(12.3567,      .00766) == "12.3567 +/- 0.0077"
    assert value_str(123.567,      .0764) == "123.567 +/- 0.076"
    assert value_str(1235.67,      .764) == "1235.67 +/- 0.76"
    assert value_str(12356.7,     7.64) == "12356.7 +/- 7.6"
    assert value_str(123567,     76.4) == "123567 +/- 76"
    assert value_str(1235670,   764) == "1.23567e6 +/- 0.00076e6"
    assert value_str(12356700,  764) == "12.35670e6 +/- 0.00076e6"
    assert value_str(123567000, 764) == "123.56700e6 +/- 0.00076e6"
    assert value_str(123567000, 7640) == "123.5670e6 +/- 0.0076e6"
    assert value_str(1235670000, 76400) == "1.235670e9 +/- 0.000076e9"

    # val_place == err_place
    assert value_str(123567, 764000) == "0.12e6 +/- 0.76e6"
    assert value_str(12356.7, 76400) == "12e3 +/- 76e3"
    assert value_str(1235.67, 7640) == "1.2e3 +/- 7.6e3"
    assert value_str(123.567, 764) == "0.12e3 +/- 0.76e3"
    assert value_str(12.3567, 76.4) == "12 +/- 76"
    assert value_str(1.23567, 7.64) == "1.2 +/- 7.6"
    assert value_str(.123567, .764) == "0.12 +/- 0.76"
    assert value_str(.0123567, .0764) == "12e-3 +/- 76e-3"
    assert value_str(.00123567, .00764) == "1.2e-3 +/- 7.6e-3"
    assert value_str(.000123567, .000764) == "0.12e-3 +/- 0.76e-3"

    # val_place == err_place-1
    assert value_str(123567, 7640000) == "0.1e6 +/- 7.6e6"
    assert value_str(12356.7, 764000) == "0.01e6 +/- 0.76e6"
    assert value_str(1235.67, 76400) == "0.001e6 +/- 0.076e6"
    assert value_str(123.567, 7640) == "0.1e3 +/- 7.6e3"
    assert value_str(12.3567, 764) == "0.01e3 +/- 0.76e3"
    assert value_str(1.23567, 76.4) == "0.001e3 +/- 0.076e3"
    assert value_str(.123567, 7.64) == "0.1 +/- 7.6"
    assert value_str(.0123567, .764) == "0.01 +/- 0.76"
    assert value_str(.00123567, .0764) == "0.001 +/- 0.076"
    assert value_str(.000123567, .00764) == "0.1e-3 +/- 7.6e-3"

    # val_place == err_place-2
    assert value_str(12356700, 7640000000) == "0.0e9 +/- 7.6e9"
    assert value_str(1235670, 764000000) == "0.00e9 +/- 0.76e9"
    assert value_str(123567, 76400000) == "0.000e9 +/- 0.076e9"
    assert value_str(12356, 7640000) == "0.0e6 +/- 7.6e6"
    assert value_str(1235, 764000) == "0.00e6 +/- 0.76e6"
    assert value_str(123, 76400) == "0.000e6 +/- 0.076e6"
    assert value_str(12, 7640) == "0.0e3 +/- 7.6e3"
    assert value_str(1, 764) == "0.00e3 +/- 0.76e3"
    assert value_str(0.1, 76.4) == "0.000e3 +/- 0.076e3"
    assert value_str(0.01, 7.64) == "0.0 +/- 7.6"
    assert value_str(0.001, 0.764) == "0.00 +/- 0.76"
    assert value_str(0.0001, 0.0764) == "0.000 +/- 0.076"
    assert value_str(0.00001, 0.00764) == "0.0e-3 +/- 7.6e-3"

    # val_place == err_place-3
    assert value_str(12356700, 76400000000) == "0.000e12 +/- 0.076e12"
    assert value_str(1235670, 7640000000) == "0.0e9 +/- 7.6e9"
    assert value_str(123567, 764000000) == "0.00e9 +/- 0.76e9"
    assert value_str(12356, 76400000) == "0.000e9 +/- 0.076e9"
    assert value_str(1235, 7640000) == "0.0e6 +/- 7.6e6"
    assert value_str(123, 764000) == "0.00e6 +/- 0.76e6"
    assert value_str(12, 76400) == "0.000e6 +/- 0.076e6"
    assert value_str(1, 7640) == "0.0e3 +/- 7.6e3"
    assert value_str(0.1, 764) == "0.00e3 +/- 0.76e3"
    assert value_str(0.01, 76.4) == "0.000e3 +/- 0.076e3"
    assert value_str(0.001, 7.64) == "0.0 +/- 7.6"
    assert value_str(0.0001, 0.764) == "0.00 +/- 0.76"
    assert value_str(0.00001, 0.0764) == "0.000 +/- 0.076"
    assert value_str(0.000001, 0.00764) == "0.0e-3 +/- 7.6e-3"

    # Zero values
    assert value_str(0, 7640000) == "0.0e6 +/- 7.6e6"
    assert value_str(0, 764000) == "0.00e6 +/- 0.76e6"
    assert value_str(0,  76400) == "0.000e6 +/- 0.076e6"
    assert value_str(0,   7640) == "0.0e3 +/- 7.6e3"
    assert value_str(0,    764) == "0.00e3 +/- 0.76e3"
    assert value_str(0,     76.4) == "0.000e3 +/- 0.076e3"
    assert value_str(0,      7.64) == "0.0 +/- 7.6"
    assert value_str(0,      0.764) == "0.00 +/- 0.76"
    assert value_str(0,      0.0764) == "0.000 +/- 0.076"
    assert value_str(0,      0.00764) == "0.0e-3 +/- 7.6e-3"
    assert value_str(0,      0.000764) == "0.00e-3 +/- 0.76e-3"
    assert value_str(0,      0.0000764) == "0.000e-3 +/- 0.076e-3"

    # negative values
    assert value_str(-1235670, 766000) == "-1.24e6 +/- 0.77e6"
    assert value_str(-1.23567, .766) == "-1.24 +/- 0.77"
    assert value_str(-.00000123567, .0000000766) == "-1.236e-6 +/- 0.077e-6"
    assert value_str(-12356.7, 7.64) == "-12356.7 +/- 7.6"
    assert value_str(-123.567, 764) == "-0.12e3 +/- 0.76e3"
    assert value_str(-1235.67, 76400) == "-0.001e6 +/- 0.076e6"
    assert value_str(-.000123567, .00764) == "-0.1e-3 +/- 7.6e-3"
    assert value_str(-12356, 7640000) == "-0.0e6 +/- 7.6e6"
    assert value_str(-12, 76400) == "-0.000e6 +/- 0.076e6"
    assert value_str(-0.0001, 0.764) == "-0.00 +/- 0.76"

    # non-finite values
    assert value_str(-inf, None) == "-inf"
    assert value_str(inf, None) == "inf"
    assert value_str(nan, None) == "NaN"

    # bad or missing uncertainty
    assert value_str(-1.23567, nan) == "-1.23567"
    assert value_str(-1.23567, -inf) == "-1.23567"
    assert value_str(-1.23567, -0.1) == "-1.23567"
    assert value_str(-1.23567, 0) == "-1.23567"
    assert value_str(-1.23567, None) == "-1.23567"
    assert value_str(-1.23567, inf) == "-1.2 +/- inf"


def test():
    # Check compact and plus/minus formats
    test_compact()
    test_pm()
    # Check that the default is the compact format
    assert format_uncertainty(-1.23567, 0.766) == "-1.24(77)"

    import doctest
    doctest.testmod()

if __name__ == "__main__":
    test()
